# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Wed Jan 07 23:37:51 2015

@author: Vidar Tonaas Fauske
"""

from hyperspyui.plugins.plugin import Plugin

from qtpy import QtCore, QtWidgets

from hyperspyui.widgets.extendedqwidgets import ExToolWindow


def tr(text):
    return QtCore.QCoreApplication.translate("Rebin", text)


class RebinPlugin(Plugin):
    name = "Rebin"

    def create_actions(self):
        self.add_action('rebin', tr("Rebin"), self.rebin_dialog,
                        icon='rebin.svg',
                        selection_callback=self.ui.select_signal,
                        tip=tr("Rebin the signal"))

    def create_menu(self):
        self.add_menuitem('Signal', self.ui.actions['rebin'])

    def rebin(self, factors, signal=None):
        if signal is None:
            signal = self.ui.get_selected_wrapper()
        s = signal.signal
        shape = []
        mods = [tuple()] * len(s.axes_manager.shape)
        for i in range(len(s.axes_manager.shape)):
            ax = s.axes_manager[i]
            factor = factors[i]
            if factor > ax.size:
                factor = ax.size
            if ax.size % factor == 0:
                mods[ax.index_in_array] = slice(None)
            else:
                mods[ax.index_in_array] = slice(None, - (ax.size % factor))
            shape.append(ax.size // factor)
        # Crop to multiple of factors
        s.data = s.data[tuple(mods)]

        # Update shape
        s.get_dimensions_from_data()

        # Do actual rebin
        signal.switch_signal(s.rebin(shape))
        self.ui.setUpdatesEnabled(False)
        try:
            signal.replot()
        finally:
            self.ui.setUpdatesEnabled(True)    # Always resume updates!
        self.record_code("<p>.rebin({0})".format(factors))

    def rebin_dialog(self, signal=None):
        if signal is None:
            signal = self.ui.get_selected_wrapper()
            if signal is None:
                return
        d = RebinDialog(signal, self, self.ui, self.ui)
        d.show()


class RebinDialog(ExToolWindow):

    def __init__(self, signal, plugin, ui, parent):
        super(RebinDialog, self).__init__(parent)
        self.signal = signal
        self.ui = ui
        self.setWindowTitle(tr("Rebin ") + signal.name)
        self.create_controls()
        self.plugin = plugin

    def rebin(self):
        factors = []
        for ax in self.signal.signal.axes_manager._get_axes_in_natural_order():
            spin = self.spins[ax.name]
            factors.append(spin.value())
        self.plugin.rebin(factors, self.signal)

    def validate(self):
        style = QtWidgets.QApplication.style()
        tmpIcon = style.standardIcon(style.SP_MessageBoxWarning, None, self)
        s = self.signal
        for ax in s.signal.axes_manager._get_axes_in_natural_order():
            spin = self.spins[ax.name]
            hbox = self.hboxes[ax.name]
            if ax.size % spin.value() != 0:
                # Not a factor, show warning
                if hbox.count() <= 1:
                    # No warning icon yet
                    iconSize = spin.height()
                    pm = tmpIcon.pixmap(iconSize, iconSize)
                    lbl = QtWidgets.QLabel()
                    lbl.setPixmap(pm)
                    lbl.setToolTip(tr("Not a factor of shape. Input data " +
                                      "will be trimmed before binning."))
                    sp = lbl.sizePolicy()
                    sp.setHorizontalPolicy(QtWidgets.QSizePolicy.Maximum)
                    lbl.setSizePolicy(sp)
                    hbox.insertWidget(0, lbl)
            else:
                # Everything OK, remove warning icon if there
                if hbox.count() > 1:
                    lbl = hbox.takeAt(0)
                    lbl.widget().deleteLater()

    def create_controls(self):
        self.spins = {}
        self.hboxes = {}
        form = QtWidgets.QFormLayout()
        for ax in self.signal.signal.axes_manager._get_axes_in_natural_order():
            spin = QtWidgets.QSpinBox(self)
            spin.setValue(1)
            spin.setMinimum(1)
            spin.valueChanged.connect(self.validate)
            self.spins[ax.name] = spin
            hbox = QtWidgets.QHBoxLayout()
            hbox.addWidget(spin)
            form.addRow(ax.name, hbox)
            self.hboxes[ax.name] = hbox
        self.form = form
        self.btn_rebin = QtWidgets.QPushButton(tr("Rebin"))
        self.btn_rebin.clicked.connect(self.rebin)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addLayout(form)
        vbox.addWidget(self.btn_rebin)

        self.setLayout(vbox)




import matplotlib
matplotlib.use('module://hyperspyui.mdi_mpl_backend')
matplotlib.interactive(True)

from hyperspyui.mdi_mpl_backend import (
    connect_on_new_figure, disconnect_on_new_figure,
    connect_on_destroy, disconnect_on_destroy)

import matplotlib.pyplot as plt


def test_new_figure(qapp):
    plt.figure()
    plt.plot(range(10))


def test_new_figure_callback(qapp):
    call_count = 0

    def trigger(figure):
        nonlocal call_count
        assert figure
        call_count += 1

    connect_on_new_figure(trigger)
    plt.figure()
    plt.plot(range(10))
    plt.close()

    disconnect_on_new_figure(trigger)
    plt.figure()
    plt.plot(range(10))
    plt.close()

    assert call_count == 1


def test_destroyed_figure_callback(qapp):
    call_count = 0

    def trigger(figure):
        nonlocal call_count
        assert figure
        call_count += 1

    connect_on_destroy(trigger)
    plt.figure()
    plt.plot(range(10))
    plt.close()

    disconnect_on_destroy(trigger)
    plt.figure()
    plt.plot(range(10))
    plt.close()

    assert call_count == 1



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Nov 23 17:10:41 2014

@author: Vidar Tonaas Fauske
"""

try:
    import traitsui.qt4.ui_base as ui_base
except RuntimeError:
    import sys

    if 'sphinx' in sys.modules:
        class Dummy(object): pass
        ui_base = Dummy()
        ui_base._StickyDialog = Dummy
    else:
        raise
orig_type = ui_base._StickyDialog


class HookedDialog(orig_type):

    def __init__(self, ui, parent, *args, **kwargs):
        _on_creating(self, ui, parent)
        super(HookedDialog, self).__init__(ui, parent, *args, **kwargs)
        _on_created(self, ui, parent)

    def closeEvent(self, e):
        _on_closing(self, e)
        super(HookedDialog, self).closeEvent(e)
        if e.isAccepted():
            _on_closed(self)

    def deleteLater(self):
        _on_destroyed(self)
        super(HookedDialog, self).deleteLater()


def hook_traitsui():
    """
    Call this function to enable hooking of traitsui events
    """
    ui_base._StickyDialog = HookedDialog


def dehook_traitsui():
    """
    Call this function to remove hooks from traitsui
    """
    ui_base._StickyDialog = orig_type

# -----------------------------------------------------------
# The rest of this file is simply event callback handling
# -----------------------------------------------------------

_creating_cbs = {}
_created_cbs = {}
_closing_cbs = {}
_closed_cbs = {}
_destroyed_cbs = {}


def _cb(cbs, *args, **kwargs):
    for cb, userdata in cbs.items():
        if userdata is None:
            cb(*args, **kwargs)
        else:
            cb(userdata, *args, **kwargs)


def _on_creating(*args, **kwargs):
    _cb(_creating_cbs, *args, **kwargs)


def _on_created(*args, **kwargs):
    _cb(_created_cbs, *args, **kwargs)


def _on_closing(*args, **kwargs):
    _cb(_closing_cbs, *args, **kwargs)


def _on_closed(*args, **kwargs):
    _cb(_closed_cbs, *args, **kwargs)


def _on_destroyed(*args, **kwargs):
    _cb(_destroyed_cbs, *args, **kwargs)


def connect_creating(callback, userdata=None):
    """
    Call to subscribe to traitsui dialog creating events. 'callback' is called
    on the event, with the 'userdata' as it's first parameter if it is not
    None; otherwise this parameter is not included in the parameter list.
    The other parameters (listed in order) are the dialog reference, traitsui's
    'ui' parameter, and the parent of the dialog. 'Creating' events are called
    just before the dialog is actually created.
    """
    global _creating_cbs
    _creating_cbs[callback] = userdata


def disconnect_creating(callback):
    """
    Disconnect callback from subscription.
    """
    global _creating_cbs
    if callback in _creating_cbs:
        _creating_cbs.pop(callback)


def connect_created(callback, userdata=None):
    """
    Call to subscribe to traitsui dialog created events. 'callback' is called
    on the event, with the 'userdata' as it's first parameter if it is not
    None; otherwise this parameter is not included in the parameter list.
    The other parameters (listed in order) are the dialog reference, traitsui's
    'ui' parameter, and the parent of the dialog. 'Created' events are called
    just after the dialog is actually created.
    """
    global _created_cbs
    _created_cbs[callback] = userdata


def disconnect_created(callback):
    """
    Disconnect callback from subscription.
    """
    global _created_cbs
    if callback in _created_cbs:
        _created_cbs.pop(callback)


def connect_closing(callback, userdata=None):
    """
    Call to subscribe to traitsui dialog closing events. 'callback' is called
    on the event, with the 'userdata' as it's first parameter if it is not
    None; otherwise this parameter is not included in the parameter list.
    The only other parameter is the dialog reference. 'Closing' events are
    called before the dialog is actually closed, and might be aborted. Is not
    fired if the dialog is deleted before it is closed. (See destroyed event)
    """
    global _closing_cbs
    _closing_cbs[callback] = userdata


def disconnect_closing(callback):
    """
    Disconnect callback from subscription.
    """
    global _closing_cbs
    if callback in _closing_cbs:
        _closing_cbs.pop(callback)


def connect_closed(callback, userdata=None):
    """
    Call to subscribe to traitsui dialog closing events. 'callback' is called
    on the event, with the 'userdata' as it's first parameter if it is not
    None; otherwise this parameter is not included in the parameter list.
    The only other parameter is the dialog reference. 'Closed' events are
    called just after the dialog is actually closed, and cannot be aborted. Is
    not fired if the dialog is deleted before it is closed. (See destroyed
    event)
    """
    global _closed_cbs
    _closed_cbs[callback] = userdata


def disconnect_closed(callback):
    """
    Disconnect callback from subscription.
    """
    global _closed_cbs
    if callback in _closed_cbs:
        _closed_cbs.pop(callback)


def connect_destroyed(callback, userdata=None):
    """
    Call to subscribe to traitsui dialog destroyed events. 'callback' is called
    on  the event, with the 'userdata' as it's first parameter if it is not
    None; otherwise this parameter is not included in the parameter list.
    The only other parameter is the dialog reference.
    """
    global _destroyed_cbs
    _destroyed_cbs[callback] = userdata


def disconnect_destroyed(callback):
    """
    Disconnect callback from subscription.
    """
    global _destroyed_cbs
    if callback in _destroyed_cbs:
        _destroyed_cbs.pop(callback)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Nov 21 19:31:50 2014

@author: Vidar Tonaas Fauske
"""

from qtpy import QtCore, QtWidgets
from qtpy.QtCore import Qt

from functools import partial
from hyperspyui._elements import elements
from .extendedqwidgets import ExClickLabel


def tr(text):
    return QtCore.QCoreApplication.translate("PeriodicTable", text)


class PeriodicTableWidget(QtWidgets.QWidget):
    element_toggled = QtCore.Signal(str)

    def __init__(self, parent=None):
        super(PeriodicTableWidget, self).__init__(parent)
        self.elements = {}
        self.toggled = {}
        self.style_off = "* { background-color: rgba(0,0,0,20); padding: 0px}"
        self.style_on = "* { background-color: rgba(128,180,255,255); padding: 0px}"
        self.style_disabled = "* { background-color: rgba(100,0,0,40); padding: 0px}"
        self.create_controls()

    def parse_elements(self, grid):
        #        btn_color = QColor(128, 128, 128)
        for i, row in enumerate(elements):
            j = 0
            for e in row:
                if isinstance(e, tuple):
                    w = QtWidgets.QLabel(e[1], self)
                    grid.addWidget(w, i, j, 1, e[0], Qt.AlignRight)
                    j += e[0]
                elif isinstance(e, dict):
                    w = ExClickLabel(e['id'], self)
                    w.setToolTip(tr(e['name'].capitalize()))
                    w.setAlignment(Qt.AlignCenter)
                    w.setMinimumSize(10, 10)

                    f = partial(self.on_element_click, e)
                    w.clicked.connect(f)

                    self.elements[e['id']] = w
                    self.toggled[e['id']] = False
                    w.setStyleSheet(self.style_off)

                    grid.addWidget(w, i, j)
                    j += 1

    def toggle_element(self, element):
        self.set_element(element, not self.toggled[element])

    def set_elements(self, elements):
        for e in self.elements.keys():
            self.set_element(e, e in elements)

    def set_element(self, element, value):
        if self.toggled[element] == value:
            return
        self.toggled[element] = value
        btn = self.elements[element]
        if value:
            style = self.style_on
        else:
            style = self.style_off
        btn.setStyleSheet(style)

    def disable_elements(self, elements):
        for e in elements:
            self.disable_element(e)

    def disable_element(self, element):
        btn = self.elements[element]
        btn.setEnabled(False)
        btn.setStyleSheet(self.style_disabled)

    def enable_elements(self, elements):
        for e in elements:
            self.enable_element(e)

    def enable_element(self, element):
        btn = self.elements[element]
        btn.setEnabled(True)
        btn.setStyleSheet(self.style_disabled)
        if self.toggled[element]:
            style = self.style_on
        else:
            style = self.style_off
        btn.setStyleSheet(style)

    def on_element_click(self, value):
        elid = value['id']
        self.set_element(elid, not self.toggled[elid])
        self.element_toggled.emit(value['id'])

    def sizeHint(self):
        return QtCore.QSize(310, 140)

    def create_controls(self):
        grid = QtWidgets.QGridLayout(self)
        grid.setSpacing(0)
        grid.setContentsMargins(0, 0, 0, 0)
        self.parse_elements(grid)



import hyperspy.signals
from hyperspyui.plugins.plugin import Plugin
from hyperspyui.util import SignalTypeFilter
from hyperspyui.widgets.signallist import SignalList
from hyperspyui.log import logger
from qtpy.QtWidgets import QDialog


class Alignzlp(Plugin):
    name = "AlignZLP"

    def create_actions(self):
        self.add_action(
            self.name +
            '.default',
            self.name,
            self.default,
            icon="align_zero_loss.svg",
            tip="Align the zero loss peak of an EELS spectrum image",
            selection_callback=SignalTypeFilter(
                hyperspy.signals.EELSSpectrum, self.ui))

    def create_menu(self):
        self.add_menuitem('EELS', self.ui.actions[self.name + '.default'])

    def create_toolbars(self):
        self.add_toolbar_button(
            'EELS', self.ui.actions[
                self.name + '.default'])

    def default(self):
        ui = self.ui
        s = ui.get_selected_signal()
        logger.debug('Align zero loss peak of ' + str(s))
        title = s.metadata.General.title

        signal_list = [
            sig for sig in ui.signals if (
                sig.signal is not s and
                isinstance(sig.signal, hyperspy.signals.EELSSpectrum))]
        also_align = []
        if len(signal_list) > 0:
            picker = SignalList(items=signal_list, parent=ui, multiselect=True)
            diag = ui.show_okcancel_dialog("Select signals to align with {}".format(
                title), picker, modal=True)
            if diag.result() != QDialog.Accepted:
                return
            also_align = [sig.signal for sig in picker.get_selected()]
        logger.debug(
            'Also aligning the following signals\n' + str(also_align))
        s.align_zero_loss_peak(also_align=also_align)
        logger.debug('ZLP alignment complete')



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri May 22 12:10:20 2015

@author: Vidar Tonaas Fauske
"""

from hyperspyui.tools import FigureTool


class SignalFigureTool(FigureTool):

    def __init__(self, windows=None):
        super(SignalFigureTool, self).__init__(windows)

    def _get_wrapper(self, figure):
        # We need to map figure to a hyperspy Signal:
        window = figure.canvas.parent()
        sw = window.property('hyperspyUI.SignalWrapper')
        return sw

    def _get_signal(self, figure):
        sw = self._get_wrapper(figure)
        if sw is None:
            return None
        return sw.signal

    def _is_nav(self, event):
        sw = self._get_wrapper(event.inaxes.figure)
        if (sw.signal._plot is not None and
                sw.signal._plot.navigator_plot is not None):
            nav_ax = sw.signal._plot.navigator_plot.ax
            return nav_ax == event.inaxes
        return False

    def _is_sig(self, event):
        sw = self._get_wrapper(event.inaxes.figure)
        if (sw.signal._plot is not None and
                sw.signal._plot.signal_plot is not None):
            sig_ax = sw.signal._plot.signal_plot.ax
            return sig_ax == event.inaxes
        return False

    def _get_axes(self, event):
        sw = self._get_wrapper(event.inaxes.figure)
        # Find out which axes of Signal are plotted in figure
        if not sw or not sw.signal:
            return
        am = sw.signal.axes_manager
        if self._is_sig(event):
            axes = am.signal_axes
        elif self._is_nav(event):
            axes = am.navigation_axes
        else:
            return
        return axes



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Oct 27 21:17:42 2014

@author: Vidar Tonaas Fauske
"""

# Set proper backend for matplotlib
import matplotlib
matplotlib.use('module://hyperspyui.mdi_mpl_backend')
matplotlib.interactive(True)

import warnings
import sys

from hyperspyui.exceptions import ProcessCanceled
from hyperspyui.log import logger
from hyperspyui.widgets.consolewidget import ConsoleWidget

from qtpy import QtCore, QtWidgets, API
from qtpy.QtCore import Qt


def myexcepthook(exctype, value, traceback):
    if exctype == ProcessCanceled:
        logger.info("User cancelled operation")
    else:
        sys.__excepthook__(exctype, value, traceback)
sys.excepthook = myexcepthook


def tr(text):
    return QtCore.QCoreApplication.translate("MainWindow", text)


def lowpriority():
    """ Set the priority of the process to below-normal."""

    if sys.platform == 'win32':
        # Based on:
        #   "Recipe 496767: Set Process Priority In Windows" on ActiveState
        #   http://code.activestate.com/recipes/496767/
        try:
            import win32api
            import win32process
            import win32con
        except ImportError as e:
            warnings.warn("Could not set process priority: %s" % e)
            return

        pid = win32api.GetCurrentProcessId()
        handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
        win32process.SetPriorityClass(
            handle, win32process.BELOW_NORMAL_PRIORITY_CLASS)
    else:
        import os

        os.nice(1)


def normalpriority():
    """ Set the priority of the process to below-normal."""

    if sys.platform == 'win32':
        # Based on:
        #   "Recipe 496767: Set Process Priority In Windows" on ActiveState
        #   http://code.activestate.com/recipes/496767/
        try:
            import win32api
            import win32process
            import win32con
        except ImportError as e:
            warnings.warn("Could not set process priority: %s" % e)
            return

        pid = win32api.GetCurrentProcessId()
        handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
        win32process.SetPriorityClass(
            handle, win32process.NORMAL_PRIORITY_CLASS)
    else:
        import os

        # Reset nice to 0
        os.nice(-os.nice(0))


class MainWindowBase(QtWidgets.QMainWindow):

    """
    Base layer in application stack. Should handle the connection to our custom
    matplotlib backend, and manage the Figures. As such, it does not know
    anything about hyperspy, and can readily be reused for other projects.
    Should also set up basic UI with utilities, and relevant functions for
    inhereting classes to override.
    """

    def __init__(self, parent=None):
        super(MainWindowBase, self).__init__(parent)

        # Do the import here to update the splash

        import hyperspyui.mdi_mpl_backend
        from hyperspyui.settings import Settings

        # Setup settings:
        self.settings = Settings(self, 'General')
        # Default setting values:
        self.settings.set_default('toolbar_button_size', 24)
        self.settings.set_default('default_widget_floating', False)
        self.settings.set_default('working_directory', "")
        self.settings.set_default('low_process_priority', False)
        # Override any possible invalid stored values, which could prevent load
        if 'toolbar_button_size' not in self.settings or \
                not isinstance(self.settings['toolbar_button_size'], int):
            self.settings['toolbar_button_size'] = 24
        if self.low_process_priority:
            lowpriority()

        # State varaibles
        self.should_capture_traits = None
        self.active_tool = None

        # Collections
        self.widgets = []   # Widgets in widget bar
        self.figures = []   # Matplotlib figures
        self.editors = []   # EditorWidgets
        self.traits_dialogs = []
        self.actions = {}
        self._action_selection_cbs = {}
        self.toolbars = {}
        self.menus = {}
        self.tools = []
        self.plugin_manager = None

        # MPL backend bindings
        hyperspyui.mdi_mpl_backend.connect_on_new_figure(self.on_new_figure)
        hyperspyui.mdi_mpl_backend.connect_on_destroy(self.on_destroy_figure)

        # Create UI
        self.windowmenu = None
        self.create_ui()

        # Connect figure management functions
        self.main_frame.subWindowActivated.connect(self.on_subwin_activated)

        # Save standard layout/state
        self.settings.set_default('_geometry', self.saveGeometry())
        self.settings.set_default('_windowState', self.saveState())

        # Restore layout/state if saved
        geometry = self.settings['_geometry']
        state = self.settings['_windowState']
        if geometry:
            self.restoreGeometry(geometry)
        if state:
            self.restoreState(state)

    @property
    def toolbar_button_size(self):
        return self.settings['toolbar_button_size', int]

    @toolbar_button_size.setter
    def toolbar_button_size(self, value):
        self.settings['toolbar_button_size'] = value
        self.setIconSize(
            QtCore.QSize(self.toolbar_button_size, self.toolbar_button_size))

    @property
    def cur_dir(self):
        return self.settings['working_directory'] or ''

    @cur_dir.setter
    def cur_dir(self, value):
        self.settings['working_directory'] = value

    @property
    def low_process_priority(self):
        return self.settings['low_process_priority', bool]

    @low_process_priority.setter
    def low_process_priority(self, value):
        self.settings['low_process_priority'] = value
        if value:
            lowpriority()
        else:
            normalpriority()

    @property
    def plugins(self):
        return self.plugin_manager.plugins

    def handleSecondInstance(self, argv):
        # overload if needed
        self.setWindowState(self.windowState() & ~Qt.WindowMinimized |
                            Qt.WindowActive)
        self.activateWindow()

    def closeEvent(self, event):
        self.settings['_geometry'] = self.saveGeometry()
        self.settings['_windowState'] = self.saveState()
        return super(MainWindowBase, self).closeEvent(event)

    def reset_geometry(self):
        self.settings.restore_key_default('_geometry')
        self.settings.restore_key_default('_windowState')
        geometry = self.settings['_geometry']
        state = self.settings['_windowState']
        if geometry:
            self.restoreGeometry(geometry)
        if state:
            self.restoreState(state)
        self.setWindowState(Qt.WindowMaximized)

    def create_ui(self):
        self.setIconSize(
            QtCore.QSize(self.toolbar_button_size, self.toolbar_button_size))
        self.main_frame = QtWidgets.QMdiArea()

        self.setCorner(Qt.TopRightCorner, Qt.RightDockWidgetArea)
        self.setCorner(Qt.TopLeftCorner, Qt.LeftDockWidgetArea)

        self.set_splash("Initializing plugins")
        self.init_plugins()

        self.set_splash("Creating default actions")
        self.create_default_actions()   # Goes before menu/toolbar/widgetbar

        # Needs to go before menu, so console can be in menu
        self.set_splash("Creating console")
        self.create_console()
        # This needs to happen before the widgetbar and toolbar
        self.set_splash("Creating menus")
        self.create_menu()
        self.set_splash("Creating toolbars")
        self.create_toolbars()
        self.set_splash("Creating widgets")
        self.create_widgetbar()

        self.setCentralWidget(self.main_frame)

    def init_plugins(self):
        from .pluginmanager import PluginManager
        self.plugin_manager = PluginManager(self)
        # Disable Version selector plugin until it is fixed
        self.plugin_manager.enabled_store['Version selector'] = False
        self.plugin_manager.init_plugins()

    def create_default_actions(self):
        """
        Create default actions that can be used for e.g. toolbars and menus,
        or triggered manually.
        """
        self.set_splash("Creating plugin actions")
        self.plugin_manager.create_actions()

        self.selectable_tools = QtWidgets.QActionGroup(self)
        self.selectable_tools.setExclusive(True)

        # Nested docking action
        ac_nested = QtWidgets.QAction(tr("Nested docking"), self)
        ac_nested.setStatusTip(tr("Allow nested widget docking"))
        ac_nested.setCheckable(True)
        ac_nested.setChecked(self.isDockNestingEnabled())
        ac_nested.triggered[bool].connect(self.setDockNestingEnabled)
        self.actions['nested_docking'] = ac_nested

        # Tile windows action
        ac_tile = QtWidgets.QAction(tr("Tile"), self)
        ac_tile.setStatusTip(tr("Arranges all figures in a tile pattern"))
        ac_tile.triggered.connect(self.main_frame.tileSubWindows)
        self.actions['tile_windows'] = ac_tile

        # Cascade windows action
        ac_cascade = QtWidgets.QAction(tr("Cascade"), self)
        ac_cascade.setStatusTip(
            tr("Arranges all figures in a cascade pattern"))
        ac_cascade.triggered.connect(self.main_frame.cascadeSubWindows)
        self.actions['cascade_windows'] = ac_cascade

        # Close all figures action
        ac_close_figs = QtWidgets.QAction(tr("Close all"), self)
        ac_close_figs.setStatusTip(tr("Closes all matplotlib figures"))
        ac_close_figs.triggered.connect(lambda: matplotlib.pyplot.close("all"))
        self.actions['close_all_windows'] = ac_close_figs

        # Reset geometry action
        ac_reset_layout = QtWidgets.QAction(tr("Reset layout"), self)
        ac_reset_layout.setStatusTip(tr("Resets layout of toolbars and "
                                        "widgets"))
        ac_reset_layout.triggered.connect(self.reset_geometry)
        self.actions['reset_layout'] = ac_reset_layout

    def create_menu(self):
        mb = self.menuBar()
        # Window menu is filled in add_widget and add_figure
        self.windowmenu = mb.addMenu(tr("&Windows"))
        self.windowmenu.addAction(self._console_dock.toggleViewAction())
        self.windowmenu.addAction(self.actions['nested_docking'])
        # Figure windows go below this separator. Other windows can be added
        # above it with insertAction(self.windowmenu_sep, QAction)
        self.windowmenu_sep = self.windowmenu.addSeparator()
        self.windowmenu.addAction(self.actions['tile_windows'])
        self.windowmenu.addAction(self.actions['cascade_windows'])
        self.windowmenu.addSeparator()
        self.windowmenu.addAction(self.actions['close_all_windows'])
        self.windowmenu_actions_sep = self.windowmenu.addSeparator()

        self.plugin_manager.create_menu()

    def create_tools(self):
        """Override to create tools on UI construction.
        """
        self.plugin_manager.create_tools()

    def create_toolbars(self):
        """
        Override to create toolbars and toolbar buttons on UI construction.
        It is called after create_default_action(), so add_toolbar_button()
        can be used to add previously defined acctions.
        """
        self.create_tools()
        self.plugin_manager.create_toolbars()

    def create_widgetbar(self):
        """
        The widget bar itself is created and managed implicitly by Qt. Override
        this function to add widgets on UI construction.
        """

        self.plugin_manager.create_widgets()

    def edit_settings(self):
        """
        Shows a dialog for editing the application and plugins settings.
        """
        from hyperspyui.widgets.settingsdialog import SettingsDialog
        d = SettingsDialog(self, self)
        d.settings_changed.connect(self.on_settings_changed)
        d.exec_()

    def on_settings_changed(self):
        """
        Callback for SettingsDialog, or anything else that updates settings
        and need to apply the change.
        """
        # Due to the way the property is defined, this updates the UI:
        self.toolbar_button_size = self.toolbar_button_size
        self.low_process_priority = self.low_process_priority

    def select_tool(self, tool):
        if self.active_tool is not None:
            try:
                self.active_tool.disconnect_windows(self.figures)
            except Exception as e:
                warnings.warn("Exception disabling tool %s: %s" % (
                    self.active_tool.get_name(), e))
        self.active_tool = tool
        tool.connect_windows(self.figures)

    # --------- Figure management ---------

    # --------- MPL Events ---------

    def on_new_figure(self, figure, userdata=None):
        """
        Callback for MPL backend.
        """
        self.main_frame.addSubWindow(figure)
        self.figures.append(figure)
        self.windowmenu.addAction(figure.activateAction())
        for tool in self.tools:
            if tool.single_action() is not None:
                tool.connect_windows(figure)
        if self.active_tool is not None:
            self.active_tool.connect_windows(figure)

    def on_destroy_figure(self, figure, userdata=None):
        """
        Callback for MPL backend.
        """
        if figure in self.figures:
            self.figures.remove(figure)
        self.windowmenu.removeAction(figure.activateAction())
        for tool in self.tools:
            if tool.single_action() is not None:
                tool.disconnect_windows(figure)
        if self.active_tool is not None:
            self.active_tool.disconnect_windows(figure)
        self.main_frame.removeSubWindow(figure)

    # --------- End MPL Events ---------

    def on_subwin_activated(self, mdi_figure):
        if mdi_figure and API == 'pyside':
            mdi_figure.activateAction().setChecked(True)
        self.check_action_selections(mdi_figure)

    def check_action_selections(self, mdi_figure=None):
        if mdi_figure is None:
            mdi_figure = self.main_frame.activeSubWindow()
        for key, cb in self._action_selection_cbs.items():
            cb(mdi_figure, self.actions[key])

    # --------- End figure management ---------


    # --------- Console functions ---------

    def _get_console_exec(self):
        return ""

    def _get_console_exports(self):
        return {'ui': self}

    def _get_console_config(self):
        return None

    def on_console_executing(self, source):
        """
        Override when inherited to perform actions before exectuing 'source'.
        """
        pass

    def on_console_executed(self, response):
        """
        Override when inherited to perform actions after executing, given the
        'response' returned.
        """
        pass

    def create_console(self):
        # We could inherit QAction, and have it reroute when it triggers,
        # and then drop route when it finishes, however this will not catch
        # interactive dialogs and such.
        c = self._get_console_config()
        self.settings.set_default('console_completion_type', 'droplist')
        valid_completions = ConsoleWidget.gui_completion.values
        self.settings.set_enum_hint('console_completion_type',
                                    valid_completions)
        gui_completion = self.settings['console_completion_type']
        if gui_completion not in valid_completions:
            gui_completion = 'droplist'
        control = ConsoleWidget(config=c, gui_completion=gui_completion)
        control.executing.connect(self.on_console_executing)
        control.executed.connect(self.on_console_executed)

        # This is where we push variables to the console
        ex = self._get_console_exec()
        push = self._get_console_exports()
        control.ex(ex)
        control.push(push)

        self.console = control

        self._console_dock = QtWidgets.QDockWidget("Console")
        self._console_dock.setObjectName('console_widget')
        self._console_dock.setWidget(control)
        self.addDockWidget(Qt.BottomDockWidgetArea, self._console_dock)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Dec 12 23:44:26 2014

@author: Vidar Tonaas Fauske
"""

from hyperspyui.settings import Settings
from functools import partial


class Plugin(object):
    name = None

    def __init__(self, main_window):
        super(Plugin, self).__init__()
        self.ui = main_window
        if self.name is None:
            set_group = 'plugins/' + str.lower(self.__class__.__name__)
            set_group = set_group.replace('.', '')
        else:
            set_group = 'plugins/' + self.name.replace('/', '-').replace(
                '\\', '-')
        self.settings = Settings(self.ui, group=set_group)

        self.actions = {}
        self.menu_actions = {}
        self.toolbar_actions = {}
        self.tools = []
        self.widgets = set()
        self.dialogs = []

    def add_action(self, key, *args, **kwargs):
        ac = self.ui.add_action(key, *args, **kwargs)
        self.actions[key] = ac

    def add_menuitem(self, category, action, *args, **kwargs):
        self.ui.add_menuitem(category, action, *args, **kwargs)
        if category in self.menu_actions:
            self.menu_actions[category].append(action)
        else:
            self.menu_actions[category] = [action]

    def add_tool(self, tool, selection_callback=None):
        self.ui.add_tool(tool, selection_callback)
        self.tools.append(tool)

    def add_toolbar_button(self, category, action, *args, **kwargs):
        self.ui.add_toolbar_button(category, action, *args, **kwargs)
        if category in self.toolbar_actions:
            self.toolbar_actions[category].append(action)
        else:
            self.toolbar_actions[category] = [action]

    def add_widget(self, widget, *args, **kwargs):
        dock = self.ui.add_widget(widget, *args, **kwargs)
        self.widgets.add(dock)

    def record_code(self, code):
        code = code.replace('<p>', "ui.plugins['{0}']".format(self.name))
        self.ui.record_code(code)

    def create_actions(self):
        pass

    def create_menu(self):
        pass

    def create_tools(self):
        pass

    def create_toolbars(self):
        pass

    def create_widgets(self):
        pass

    def unload(self):
        for category, actions in self.menu_actions.items():
            for action in actions:
                self.ui.menus[category].removeAction(action)
        for category, actions in self.toolbar_actions.items():
            for action in actions:
                self.ui.toolbars[category].removeAction(action)
        for key in self.actions.keys():
            self.ui.actions.pop(key, None)
        for tool in self.tools:
            self.ui.remove_tool(tool)
        for widget in self.widgets:
            self.ui.removeDockWidget(widget)
            if widget in self.ui.widgets:
                self.ui.widgets.remove(widget)
            else:
                # Our widget was wrapped, original is in ui.widgets.
                self.ui.widgets.remove(widget.widget())
        self.actions.clear()
        self.menu_actions.clear()
        self.toolbar_actions.clear()
        self.widgets.clear()

    def _on_dialog_close(self, diag):
        self.dialogs.remove(diag)
        diag.deleteLater()

    def open_dialog(self, dialog, modal=False, delete_on_close=True):
        self.dialogs.append(dialog)
        if delete_on_close:
            dialog.finished.connect(partial(self._on_dialog_close, dialog))
        if modal:
            dialog.exec_()
        else:
            dialog.show()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Wed Nov 26 19:11:19 2014

@author: Vidar Tonaas Fauske
"""

from __future__ import division, absolute_import
# future division is important to divide integers and get as
# a result precise floating numbers (instead of truncated int)
# import compatibility functions and utilities
import sys
from time import time

from qtpy.QtCore import QObject, Signal

import hyperspy.external.progressbar
from tqdm import tqdm

from hyperspyui.exceptions import ProcessCanceled


# Create signal object which will handle all events
class Signaler(QObject):
    created = Signal([int, int, str])
    progress = Signal([int, int, str], [int, int])
    finished = Signal(int)

    cancel = Signal(int)
    # This is necessary as it bugs out if not (it's a daisy chained event)

    def _on_cancel(self, pid):
        signaler.cancel.emit(pid)
    on_cancel = _on_cancel


signaler = Signaler()

# Hook function


def _wrap(*args, **kwargs):
    """
    Replacement function for hyperspy.external.progressbar.progressbar().
    Causes a UIProgressBar() to be made, which the MainWindow can connect to
    in order to create a progress indicator. It is important that the
    connection is made with QtCore.Signals, as they are thread aware, and the
    signal is processed on the GUI main event loop, i.e. the main thread. This
    is necessary as all UI operations have to happen on the main thread, and
    the hyperspy processing might be pushed to a worker thread "threaded.py".
    """
    return UIProgressBar(*args, **kwargs)


# Override hyperspy prgoressbar implementation
orig = hyperspy.external.progressbar.progressbar


def takeover_progressbar():
    """
    Replace hyperspy.external.progressbar.progressbar() with uiprogressbar.wrap().
    The main_window will be connected to all the events whenever a progressbar
    is created.
    """
    hyperspy.external.progressbar.progressbar = _wrap


def reset_progressbar():
    hyperspy.external.progressbar.progressbar = orig


class UIProgressBar(tqdm):

    """
    Connector between hyperspy process with a progressbar, and the UI. See also
    the doc for wrap() for more details.
    """

    uid = 1

    @classmethod
    def write(cls, s, file=sys.stdout, end="\n"):
        """
        Print a message via tqdm_gui (just an alias for print)
        """
        # TODO: print text on GUI?
        file.write(s)
        file.write(end)

    def __init__(self, *args, **kwargs):
        self.id = self.uid
        self.uid += 1
        kwargs['gui'] = True
        self.cancelled = False
        super().__init__(*args, **kwargs)
        # Initialize the GUI display
        if self.disable or not kwargs['gui']:
            return

        self.mininterval = max(self.mininterval, 0.5)

        # assert maxval >= 0
        # self.maxval = maxval
        self.signal_set = False

        global signaler
        signaler.cancel[int].connect(self.cancel)

        self.currval = 0
        self.finished = False
        self.start_time = None
        self.seconds_elapsed = 0

        signaler.created[int, int, str].emit(self.id, self.total, "")

    def cancel(self, pid):
        """
        Slot for the UI to call if it wants to cancel the process. Thread safe.
        """
        if pid == self.id:
            self.cancelled = True

    @staticmethod
    def format_string(n, total, elapsed, rate=None):
        return "ETA: " + (tqdm.format_interval((total - n) / rate)
                          if rate else '?')

    def __iter__(self):
        iterable = self.iterable
        if self.disable:
            for obj in iterable:
                if self.cancelled is True:
                    raise ProcessCanceled("User cancelled operation")
                yield obj
            return

        # ncols = self.ncols
        mininterval = self.mininterval
        maxinterval = self.maxinterval
        miniters = self.miniters
        dynamic_miniters = self.dynamic_miniters
        start_t = self.start_t
        last_print_t = self.last_print_t
        last_print_n = self.last_print_n
        n = self.n
        # dynamic_ncols = self.dynamic_ncols
        smoothing = self.smoothing
        avg_time = self.avg_time

        for obj in iterable:
            if self.cancelled is True:
                raise ProcessCanceled("User cancelled operation")
            yield obj
            # Update and print the progressbar.
            # Note: does not call self.update(1) for speed optimisation.
            n += 1
            delta_it = n - last_print_n
            # check the counter first (avoid calls to time())
            if delta_it >= miniters:
                cur_t = time()
                delta_t = cur_t - last_print_t
                if delta_t >= mininterval:
                    elapsed = cur_t - start_t
                    # EMA (not just overall average)
                    if smoothing and delta_t:
                        avg_time = delta_t / delta_it \
                            if avg_time is None \
                            else smoothing * delta_t / delta_it + \
                            (1 - smoothing) * avg_time

                    txt = self.format_string(
                        n, self.total, elapsed,
                        1 / avg_time if avg_time else None)

                    global signaler
                    signaler.progress[int, int, str].emit(self.id, n, txt)

                    # If no `miniters` was specified, adjust automatically
                    # to the maximum iteration rate seen so far.
                    if dynamic_miniters:
                        if maxinterval and delta_t > maxinterval:
                            # Set miniters to correspond to maxinterval
                            miniters = delta_it * maxinterval / delta_t
                        elif mininterval and delta_t:
                            # EMA-weight miniters to converge
                            # towards the timeframe of mininterval
                            miniters = smoothing * delta_it * mininterval \
                                / delta_t + (1 - smoothing) * miniters
                        else:
                            miniters = smoothing * delta_it + \
                                (1 - smoothing) * miniters

                    # Store old values for next call
                    last_print_n = n
                    last_print_t = cur_t

        # Closing the progress bar.
        # Update some internal variables for close().
        self.last_print_n = last_print_n
        self.n = n
        self.close()

    def update(self, n=1):
        """
        Updates the progress bar to a new value. Called by the hyperspy side.
        Not safe to call from UI.
        """
        if self.disable:
            return
        if self.cancelled is True:
            raise ProcessCanceled("User cancelled operation")

        if n < 0:
            n = 1
        self.n += n

        delta_it = self.n - self.last_print_n  # should be n?
        if delta_it >= self.miniters:
            # We check the counter first, to reduce the overhead of time()
            cur_t = time()
            delta_t = cur_t - self.last_print_t
            if delta_t >= self.mininterval:
                elapsed = cur_t - self.start_t
                # EMA (not just overall average)
                if self.smoothing and delta_t:
                    self.avg_time = delta_t / delta_it \
                        if self.avg_time is None \
                        else self.smoothing * delta_t / delta_it + \
                        (1 - self.smoothing) * self.avg_time

                txt = self.format_string(
                    self.n, self.total, elapsed,
                    1 / self.avg_time if self.avg_time else None)

                global signaler
                signaler.progress[int, int, str].emit(self.id, self.n, txt)

                # If no `miniters` was specified, adjust automatically to the
                # maximum iteration rate seen so far.
                # e.g.: After running `tqdm.update(5)`, subsequent
                # calls to `tqdm.update()` will only cause an update after
                # at least 5 more iterations.
                if self.dynamic_miniters:
                    if self.maxinterval and delta_t > self.maxinterval:
                        self.miniters = self.miniters * self.maxinterval \
                            / delta_t
                    elif self.mininterval and delta_t:
                        self.miniters = self.smoothing * delta_it \
                            * self.mininterval / delta_t + \
                            (1 - self.smoothing) * self.miniters
                    else:
                        self.miniters = self.smoothing * delta_it + \
                            (1 - self.smoothing) * self.miniters

                # Store old values for next call
                self.last_print_n = self.n
                self.last_print_t = cur_t

    def close(self):
        if self.disable:
            return

        self.disable = True
        self.finish()

        self._instances.remove(self)

    def finish(self):
        """
        Used to tell the progress is finished. Called by hyperspy side.
        """
        global signaler
        signaler.finished[int].emit(self.id)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Dec 12 23:57:20 2014

@author: Vidar Tonaas Fauske
"""

import os
import glob
modules = glob.glob(os.path.dirname(__file__) + "/*.py")
# modules.extend(glob.glob(os.path.dirname(__file__)+"/*.py?"))
# TODO: In release form, we should support compiled plugins in pyc/pyo format
__all__ = [os.path.splitext(os.path.basename(f))[0] for f in modules
           if not os.path.basename(f).startswith('_')]
if os.path.exists(os.path.join(os.path.dirname(__file__), "user_plugins")):
    modules = glob.glob(os.path.dirname(__file__) + "/user_plugins/*.py")
    __all__.extend(['.'.join(('user_plugins',
                             os.path.splitext(os.path.basename(f))[0]))
                   for f in modules
                   if not os.path.basename(f).startswith('_')])
__all__ = list(set(__all__))    # Make unique as py/pyc/pyo all match



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Nov 17 11:35:52 2014

@author: Vidar Tonaas Fauske
"""

from qtpy import QtCore, QtWidgets
from collections import OrderedDict


class Actionable(QtCore.QObject):

    """
    Base class for simple action management utilities. Manages actions dict,
    and adds actions through add_action().
    """

    def __init__(self):
        super(Actionable, self).__init__()
        self.actions = OrderedDict()
        self.sep_counter = 0

    def add_action(self, key, title, on_trig):
        ac = QtWidgets.QAction(title, self)  # TODO: tr()?
        ac.triggered.connect(on_trig)
        self.actions[key] = ac

    def add_separator(self):
        self.sep_counter += 1
        ac = QtWidgets.QAction(self)
        ac.setSeparator(True)
        self.actions[self.sep_counter] = ac



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sat Dec 27 14:21:00 2014

@author: Vidar Tonaas Fauske
"""

from qtpy.QtCore import QSettings
from qtpy.QtWidgets import QMessageBox

from hyperspyui.widgets.extendedqwidgets import ExRememberPrompt


class Settings(object):

    def __init__(self, parent=None, group=None):
        self._sep = '/'
        self.group = group
        self.parent = parent

    def _get_groups(self, key):
        if self.group is None:
            return key.split(self._sep)
        else:
            return (self.group + self._sep + key).split(self._sep)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            key, t = key
        else:
            t = None
        groupings = self._get_groups(key)
        if key not in self:
            groupings.insert(0, 'defaults')
        key = groupings.pop()
        settings = QSettings(parent=self.parent)
        for g in groupings:
            settings.beginGroup(g)
        ret = settings.value(key)
        if t and isinstance(t, type) and not isinstance(ret, t):
            if t is bool:
                ret = ("true" == ret.lower())
            else:
                ret = t(ret)
        for g in groupings:
            settings.endGroup()
        return ret

    def __setitem__(self, key, value):
        groupings = self._get_groups(key)
        key = groupings.pop()
        settings = QSettings(parent=self.parent)
        for g in groupings:
            settings.beginGroup(g)
        settings.setValue(key, value)
        for g in groupings:
            settings.endGroup()

    def __contains__(self, key):
        groupings = self._get_groups(key)
        key = groupings.pop()
        settings = QSettings(parent=self.parent)
        for g in groupings:
            settings.beginGroup(g)
        r = settings.contains(key)
        for g in groupings:
            settings.endGroup()
        return r

    def __iter__(self):
        settings = QSettings(parent=self.parent)
        settings.beginGroup(self.group)
        keys = settings.allKeys()
        for k in keys:
            yield k, settings.value(k)

    @staticmethod
    def clear_defaults():
        """
        Clear all settings in defaults group. Should only be run once during
        application start, as it will undo any defaults that have been set.
        """
        settings = QSettings()
        settings.beginGroup('defaults')
        settings.remove("")
        settings.endGroup()

    @staticmethod
    def restore_from_defaults():
        """
        Clears all settings (except "defaults" group) and restores all settings
        from the defaults group.
        """
        settings = QSettings()
        for g in settings.childGroups():
            if g != "defaults":
                settings.remove(g)
        for k in settings.childKeys():
            settings.remove(k)
        defaults = QSettings()
        defaults.beginGroup("defaults")
        for k in defaults.allKeys():
            settings.setValue(k, defaults.value(k))

    def restore_key_default(self, key):
        """
        Restore a given setting to its default value.
        """
        groupings = self._get_groups(key)
        groupings.insert(0, 'defaults')
        inner_key = groupings.pop()
        settings = QSettings(parent=self.parent)
        for g in groupings:
            settings.beginGroup(g)
        default_value = settings.value(inner_key)
        for g in groupings:
            settings.endGroup()
        self[key] = default_value

    def set_default(self, key, value):
        """
        Sets default value by writing into defaults group.
        """
        # If not in normal settings, set it:
        if key not in self:
            self[key] = value

        # Either way, write to defaults group
        groupings = self._get_groups(key)
        groupings.insert(0, 'defaults')
        key = groupings.pop()
        settings = QSettings(parent=self.parent)
        for g in groupings:
            settings.beginGroup(g)
        settings.setValue(key, value)
        for g in groupings:
            settings.endGroup()

    def set_enum_hint(self, key, options):
        """
        Indicate possible values for a setting.

        The `options` are not strictly enforced, but can be used to indicate
        valid values to the user. A typical usecase is to allow the use of a
        combobox in a dialog to pick a value.
        """
        groupings = self._get_groups(key)
        groupings.insert(0, 'defaults')
        key = groupings.pop()
        key = '_' + key + '_options'    # Change key to avoid conflicts
        settings = QSettings(parent=self.parent)
        for g in groupings:
            settings.beginGroup(g)
        if not isinstance(options, list):
            options = list(options)
        settings.setValue(key, options)
        for g in groupings:
            settings.endGroup()

    def get_enum_hint(self, key):
        """
        Returns the possible enum hint values if set, otherwise None.
        """
        groupings = self._get_groups(key)
        groupings.insert(0, 'defaults')
        key = groupings.pop()
        key = '_' + key + '_options'    # Change key to avoid conflicts
        settings = QSettings(parent=self.parent)
        for g in groupings:
            settings.beginGroup(g)
        value = settings.value(key)
        for g in groupings:
            settings.endGroup()
        return value

    def get_or_prompt(self, key, options, title="Prompt", descr=""):
        """
        Gets the setting specified by key. If it is not set, prompts the user
        to select one option out of several. The prompt includes a checkbox to
        remember the answer ("Remember this setting").

        The option parameter should be a list of two-tuples, specifying an
        ordered list of option values, and labels.
        """

        # First check if we have a remembered setting.
        val = self[key]
        if val is not None:
            return val

        # Setup the dialog
        mb = ExRememberPrompt(QMessageBox.Question, title, descr)
        if len(options) < 5:
            buttons = []
            opt = options[0]
            buttons.append(mb.addButton(opt[1], QMessageBox.AcceptRole))
            for opt in options[1:]:
                buttons.append(mb.addButton(opt[1], QMessageBox.RejectRole))
        else:
            pass  # TODO: Make list selection
        mb.addButton(QMessageBox.Cancel)

        # Show the dialog modal/blocking
        mb.exec_()
        btn = mb.clickedButton()
        if btn not in buttons:
            # The user did not make a valid selection = cancelled
            return None
        sel = btn.text()
        idx = [o[1] for o in options].index(sel)
        ret = options[idx][0]
        if mb.isChecked():
            self[key] = ret
        return ret

    def write(self, d, group=None, settings=None):
        if settings is None:
            settings = QSettings(self)
        if group is not None:
            settings.beginGroup(group)

        for k, v in d.items():
            settings.setValue(k, v)

        if group is not None:
            settings.endGroup()

    def read(self, d, group=None, settings=None):
        if settings is None:
            settings = QSettings(self)
        if group is not None:
            settings.beginGroup(group)

        for k, v in d.items():
            if isinstance(v, tuple):
                settings.value(k, v)

        if group is not None:
            settings.endGroup()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Wed Jan 20 22:32:58 2016

@author: Vidar Tonaas Fauske
"""

import numpy as np
import matplotlib.pyplot as plt

from hyperspyui.plugins.plugin import Plugin
from hyperspyui.widgets.extendedqwidgets import FigureWidget
from hyperspyui.util import win2fig, fig2image_plot

from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtCore import Qt


def tr(text):
    return QtCore.QCoreApplication.translate("ColorMapPickerPlugi", text)


class CMapPickerPlugin(Plugin):
    name = "Color map picker"

    def create_widgets(self):
        w = CMapPickerWidget(self.ui, self.ui)
        w.plugin = self
        self.add_widget(w)
        # w.hide()  # Initial state hidden

    def change_cmap(self, cmap, mpl_axes=None):
        pass


# Have colormaps separated into categories:
# http://matplotlib.org/examples/color/colormaps_reference.html
cmaps = [('Uniform',        ['viridis', 'inferno', 'plasma', 'magma']),
         ('Sequential',     ['Blues', 'BuGn', 'BuPu',
                             'GnBu', 'Greens', 'Greys', 'Oranges', 'OrRd',
                             'PuBu', 'PuBuGn', 'PuRd', 'Purples', 'RdPu',
                             'Reds', 'YlGn', 'YlGnBu', 'YlOrBr', 'YlOrRd']),
         ('Sequential (2)', ['afmhot', 'autumn', 'bone', 'cool',
                             'copper', 'gist_heat', 'gray', 'hot',
                             'pink', 'spring', 'summer', 'winter']),
         ('Diverging',      ['BrBG', 'bwr', 'coolwarm', 'PiYG', 'PRGn', 'PuOr',
                             'RdBu', 'RdGy', 'RdYlBu', 'RdYlGn', 'Spectral',
                             'seismic']),
         ('Qualitative',    ['Accent', 'Dark2', 'Paired', 'Pastel1',
                             'Pastel2', 'Set1', 'Set2', 'Set3']),
         ('Miscellaneous',  ['gist_earth', 'terrain', 'ocean', 'gist_stern',
                             'brg', 'CMRmap', 'cubehelix',
                             'gnuplot', 'gnuplot2', 'gist_ncar',
                             'nipy_spectral', 'jet', 'rainbow',
                             'gist_rainbow', 'hsv', 'flag', 'prism'])]


class CMapDelegate(QtWidgets.QItemDelegate):
    """
    Delegate responsible for drawing a the entries in the colormap combobox.

    In addition to drawing a preview of the colormap, it also allows for
    parent elements to be added (group headings). The colormaps will be
    prepended by a set amount to prevent overlapping with the text.
    """

    def __init__(self, width, height, cmap_shift, *args, **kwargs):
        super(CMapDelegate, self).__init__(*args, **kwargs)
        self.width_ = width
        self.height_ = height
        self.cmap_shift = cmap_shift

    @staticmethod
    def _create_pixmap(line, height):
        width = line.shape[0]
        data = line[:, [2, 1, 0, 3]]
        data = np.tile(data, (height, 1, 1))
        im = QtGui.QImage(data.data, width, height, QtGui.QImage.Format_ARGB32)
        im.ndarray = data
        return QtGui.QPixmap.fromImage(im)

    def paint(self, painter, option, index):

        kind = index.data(Qt.AccessibleDescriptionRole)
        if kind == "parent":
            option.state |= QtWidgets.QStyle.State_Enabled
        else:
            option.textElideMode = Qt.ElideNone
            rect = option.rect
            rect = QtCore.QRect(rect.x()+self.cmap_shift, rect.y(),
                                rect.width()-self.cmap_shift, rect.height())
            option.rect.setLeft(option.rect.x() + 10)
        super(CMapDelegate, self).paint(painter, option, index)

        if kind != "parent":
            cmap = plt.get_cmap(index.data().strip())
            width = self.width_ - self.cmap_shift
            line = cmap(np.linspace(0, 1, width, endpoint=True))
            line = (255 * line).astype(np.uint8)
            pixmap = self._create_pixmap(line, self.height_)
            painter.drawPixmap(rect, pixmap)

    def sizeHint(self, option, index):
        return QtCore.QSize(self.width_, self.height_)

    def drawFocus(self, painter, option, rect):
        super(CMapDelegate, self).drawFocus(painter, option, rect)


class CMapPickerWidget(FigureWidget):

    def __init__(self, main_window, parent, figure=None):
        super(CMapPickerWidget, self).__init__(main_window, parent)
        self.setWindowTitle(tr("Colormap Picker"))
        self.create_controls()

        self._on_figure_change(figure)

    def _on_select(self, cmap):
        img = self._cur_img
        if img is not None:
            for im in img:
                im.set_cmap(cmap)
            im.figure.canvas.draw()

    def _on_figure_change(self, figure):
        super(CMapPickerWidget, self)._on_figure_change(figure)
        if isinstance(figure, QtWidgets.QMdiSubWindow):
            figure = win2fig(figure)

        signals = self.parent().signals
        p = fig2image_plot(figure, signals)
        img = None
        if p is not None and len(p.ax.images) > 0:
            img = p.ax.images

        self._cur_img = img
        self.update_controls_from_fig()

    def update_controls_from_fig(self):
        img = self._cur_img
        if img is None:
            self.disable()
        else:
            self.enable()
            cmap = img[0].get_cmap().name
            self.cbo.setCurrentIndex(self.cbo.findText(cmap))

    def enable(self, enabled=True):
        self.cbo.setEnabled(enabled)

    def disable(self):
        self.enable(False)

    @staticmethod
    def make_parent_item(text):
        item = QtGui.QStandardItem(text)
        item.setFlags(item.flags() & ~(Qt.ItemIsEnabled | Qt.ItemIsSelectable))
        item.setData("parent", Qt.AccessibleDescriptionRole)

        font = item.font()
        font.setBold(True)
        item.setFont(font)

        return item

    def create_controls(self):
        self.cbo = QtWidgets.QComboBox()
        self.cbo.setItemDelegate(CMapDelegate(256, 18, 80, self.cbo))
        for group, maps in cmaps:
            self.cbo.model().appendRow(self.make_parent_item(group))
            for cm in maps:
                self.cbo.addItem(cm)
        self.cbo.currentIndexChanged[str].connect(
            self._on_select)
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.cbo)

        wrap = QtWidgets.QWidget()
        wrap.setLayout(vbox)
        height = vbox.sizeHint().height()
        wrap.setFixedHeight(height)
        self.setWidget(wrap)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Feb 22 18:46:40 2015

@author: Vidar Tonaas Fauske
"""

from hyperspyui.plugins.plugin import Plugin

from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import (QPushButton, QVBoxLayout, QWidget, QHBoxLayout,
                            QCheckBox)

from hyperspyui.recorder import Recorder
from hyperspyui.widgets.editorwidget import EditorWidget


def tr(text):
    return QtCore.QCoreApplication.translate("RecorderWidget", text)


class RecorderWidget(QtWidgets.QDockWidget):

    def __init__(self, main_window, parent=None):
        super(RecorderWidget, self).__init__(parent)
        self.setWindowTitle(tr("Recorder"))
        self.ui = main_window
        self.create_controls()
        self.recorder = None

    def start_recording(self):
        self.btn_start.setEnabled(False)
        self.btn_stop.setEnabled(True)
        r = Recorder()
        self.recorder = r
        self.update_filter()
        self.ui.recorders.append(r)
        e = EditorWidget(self.ui, self.ui)
        e.setWindowTitle(tr("Recorded Code"))
        self.editor = e
        self.ui.editors.append(e)
        r.record.connect(e.append_code)
        self.editor.finished.connect(lambda: self.ui.editors.remove(e))
        self.editor.finished.connect(lambda: self.disconnect_editor(e))
        e.show()

    def stop_recording(self, disconnect=True):
        self.btn_start.setEnabled(True)
        self.btn_stop.setEnabled(False)
        self.ui.recorders.remove(self.recorder)
        if disconnect:
            self.disconnect_editor()
        self.recorder = None
        self.editor = None

    def update_filter(self):
        if self.recorder is None:
            return
        self.recorder.filter['actions'] = self.chk_actions.isChecked()
        self.recorder.filter['code'] = self.chk_code.isChecked()

    def disconnect_editor(self, editor=None):
        if editor is None:
            editor = self.editor
        else:
            if self.editor != editor:
                return
            self.stop_recording(disconnect=False)
        if self.recorder is not None:
            self.recorder.record.disconnect(editor.append_code)

    def create_controls(self):
        self.btn_start = QPushButton(tr("Start"))
        self.btn_stop = QPushButton(tr("Stop"))
        self.btn_stop.setEnabled(False)

        self.btn_start.clicked.connect(self.start_recording)
        self.btn_stop.clicked.connect(self.stop_recording)

        self.chk_actions = QCheckBox(tr("Actions"))
        self.chk_code = QCheckBox(tr("Code"))
        for c in [self.chk_actions, self.chk_code]:
            c.setChecked(True)
            c.toggled.connect(self.update_filter)

        hbox = QHBoxLayout()
        for w in [self.btn_start, self.btn_stop]:
            hbox.addWidget(w)
        hbox2 = QHBoxLayout()
        for w in [self.chk_actions, self.chk_code]:
            hbox2.addWidget(w)

        vbox = QVBoxLayout()
        vbox.addLayout(hbox)
        vbox.addLayout(hbox2)

        wrap = QWidget()
        wrap.setLayout(vbox)
        height = vbox.sizeHint().height()
        wrap.setFixedHeight(height)
        self.setWidget(wrap)


class RecorderWidgetPlugin(Plugin):
    name = "Recorder Widget"

    def create_widgets(self):
        self.add_widget(RecorderWidget(self.ui, self.ui))



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys

_console_mode_header = """from hyperspyui.mainwindow import MainWindow
import numpy as np
import hyperspy.api as hs
ui = MainWindow()
siglist = [hs.signals.Signal(None), hs.signals.Signal(None)]

"""
_header_num_lines = _console_mode_header.count('\n')

if __name__ == '__main__':
    from pyqode.python.backend.workers import JediCompletionProvider
    import jedi.api
    jedi.api.preload_module(["hyperspyui.mainwindow", "numpy",
                             "hyperspy.api"])

    class ConsoleJediCompletionProvider():
        @staticmethod
        def complete(code, line, column, path, encoding, prefix):
            code = _console_mode_header + code
            return JediCompletionProvider.complete(code,
                                                   line + _header_num_lines,
                                                   column, path, encoding,
                                                   prefix)

    """
    Server process' entry point
    """
    # setup argument parser and parse command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("port", help="the local tcp port to use to run "
                        "the server")
    parser.add_argument('-s', '--syspath', nargs='*')
    args = parser.parse_args()

    # add user paths to sys.path
    if args.syspath:
        for path in args.syspath:
            print('append path %s to sys.path' % path)
            sys.path.append(path)

    from pyqode.core import backend

    # setup completion providers
    backend.CodeCompletionWorker.providers.append(
            ConsoleJediCompletionProvider())
    backend.CodeCompletionWorker.providers.append(
        backend.DocumentWordsProvider())

    # starts the server
    backend.serve_forever(args)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Aug 21 19:59:04 2016

@author: Vidar Tonaas Fauske
"""


from qtpy import QtCore, QtWidgets


# QAction.trigger/activate are not virtual, so cannot simply override.
# Instead, we shadow/wrap the triggered signal used by our python code
class AdvancedAction(QtWidgets.QAction):
    # The overloaded signal can take an optional argument `advanced`
    _triggered = QtCore.Signal([], [bool], [bool, bool])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # not working on pyside, not tested on pyside2 because of pyface issue
        super().triggered[bool].connect(self._trigger)

    @property
    def triggered(self):
        return self._triggered

    def _trigger(self, checked):
        mods = QtWidgets.QApplication.keyboardModifiers()
        advanced = mods & QtCore.Qt.AltModifier == QtCore.Qt.AltModifier
        self._triggered[bool, bool].emit(checked, advanced)
        self._triggered[bool].emit(checked)
        self._triggered.emit()

    def setIcon(self, icon):
        # We need to keep a reference to the icon object if not it will be
        # garbage collected
        # See https://www.riverbankcomputing.com/pipermail/pyqt/2019-March/041459.html
        self._icon = icon
        super().setIcon(icon)

AdvancedAction.__init__.__doc__ = QtWidgets.QAction.__init__.__doc__



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Aug 03 19:43:52 2015

@author: Vidar Tonaas Fauske
"""

from qtpy import QtCore

from hyperspy.signal import BaseSignal
from hyperspy.drawing.widgets import (RectangleWidget, RangeWidget,
                                      SquareWidget, VerticalLineWidget)
from hyperspy.roi import RectangularROI, SpanROI, Point1DROI, Point2DROI

from hyperspyui.tools import SignalFigureTool
from hyperspyui.util import crosshair_cursor
import copy


class MultiSelectionTool(SignalFigureTool):

    """
    Tool to select multiple ROIs on a  signal interactively. Simply click and
    drag in a figure to create an ROI. This can be done multiple times. Right-
    click the ROI to remove it. Press enter to indicate that the selection is
    complete, or ESC to cancel. The selection can also be aborted by simply
    selecting a different tool.
    """

    accepted = QtCore.Signal([BaseSignal, list],
                             [BaseSignal, list, SignalFigureTool])
    updated = QtCore.Signal([BaseSignal, list],
                            [BaseSignal, list, SignalFigureTool])
    cancelled = QtCore.Signal([BaseSignal])

    def __init__(self, windows=None):
        super(MultiSelectionTool, self).__init__(windows)
        self.widgets = {}
        self.axes = {}
        self.valid_dimensions = [1, 2]
        self.ranged = True
        self.name = "Multi-selection tool"
        self.category = 'Signal'
        self.icon = None
        self.validator = self._default_validator

    def ndim(self, signal):
        if self.axes is None or signal is None or signal not in self.axes:
            return 0
        return len(self.axes[signal])

    def get_name(self):
        return self.name

    def get_category(self):
        return self.category

    def make_cursor(self):
        return crosshair_cursor()

    def get_icon(self):
        return self.icon

    def is_selectable(self):
        return True

    def _add_widget(self, signal):
        if signal is None:
            return None
        if self.ndim(signal) == 1:
            if self.ranged:
                w = RangeWidget(None)
            else:
                w = VerticalLineWidget(None)
        else:
            if self.ranged:
                w = RectangleWidget(None)
            else:
                w = SquareWidget(None)
        if signal in self.widgets:
            self.widgets[signal].append(w)
        else:
            self.widgets[signal] = [w]
        w.set_on(False)
        return w

    def _remove_widget(self, widget, signal=None):
        if widget.is_on():
            widget.set_on(False)
        if signal is None:
            for widgets in self.widgets.values():
                if widget in widgets:
                    widgets.remove(widget)
        else:
            self.widgets[signal].remove(widget)

    def have_selection(self, signal):
        if signal is None:
            return False
        if signal not in self.widgets:
            return False
        for w in self.widgets[signal]:
            if w.is_on():
                return True
        return False

    def on_keyup(self, event):
        if event.inaxes is None:
            s = None
        else:
            s = self._get_signal(event.inaxes.figure)
        if event.key == 'enter':
            self.accept(s)
        elif event.key == 'escape':
            self.cancel(s)

    def on_mouseup(self, event):
        # Only accept mouse up inside axes
        if event.inaxes is None:
            return

        if event.button != 3:
            # Don't handle anything except right click up events.
            return

        s = self._get_signal(event.inaxes.figure)
        for w in self.widgets[s]:
            if any([p.contains(event)[0] == True for p in w.patch]):
                self._remove_widget(w, s)
                self._on_change(w, s)

    def _default_validator(self, signal, axes):
        # Validate signal
        if signal is None:
            return False

        # Make sure we have a figure with valid dimensions
        return len(axes) in self.valid_dimensions

    def on_mousedown(self, event):
        # Only accept mouse down inside axes
        if event.inaxes is None:
            return

        if event.button == 3:
            # Don't handle right click down events.
            return

        # Find out which axes of Signal are plotted in figure
        s = self._get_signal(event.inaxes.figure)
        axes = self._get_axes(event)
        if not self.validator(s, axes):
            return

        self.axes[s] = axes

        # If we already have widgets, make sure editing is passed through
        if self.have_selection(s):
            for w in self.widgets[s]:
                if any([p.contains(event)[0] == True for p in w.patch]):
                    return              # Moving, handle in widget
            # Clicked outside existing widget, check for resize handles
            if self.ndim(s) > 1:
                for w in self.widgets[s]:
                    if w.resizers:
                        for r in w._resizer_handles:
                            if r.contains(event)[0] == True:
                                return      # Leave the event to widget
        am = s.axes_manager

        # Start a new widget
        widget = self._add_widget(s)
        widget.axes = axes
        if self.ndim(s) > 1:
            widget.axes_manager = am

        # Have what we need, create widget
        x, y = event.xdata, event.ydata
        widget.axes = axes
        widget.set_mpl_ax(event.inaxes)  # connects
        if self.ndim(s) == 1:
            widget.position = (x,)
            widget.size = axes[0].scale
            widget.set_on(True)
            if self.ranged:
                span = widget.span
                span.buttonDown = True
                span.on_move_cid = \
                    span.canvas.mpl_connect('motion_notify_event',
                                            span.move_right)
            else:
                widget.picked = True
        else:
            widget.position = (x, y)
            widget.size = [ax.scale for ax in axes]
            widget.set_on(True)
            if self.ranged:
                widget.resizer_picked = 3
            widget.picked = True
        widget.events.changed.connect(self._on_change, {'obj': 'widget'})

    def _get_rois(self, signal):
        rois = []
        if self.have_selection(signal):
            if self.ndim(signal) == 1:
                if self.ranged:
                    roi_template = SpanROI(0, 1)
                else:
                    roi_template = Point1DROI(0)
            elif self.ndim(signal) > 1:
                if self.ranged:
                    roi_template = RectangularROI(0, 0, 1, 1)
                else:
                    roi_template = Point2DROI(0, 0)
            for w in self.widgets[signal]:
                roi = copy.deepcopy(roi_template)
                # ROI gets coords from widget:
                roi._on_widget_change(w)
                rois.append(roi)
        return rois

    def _on_change(self, widget, signal=None):
        if signal is None:
            for s, widgets in self.widgets.items():
                if widget in widgets:
                    signal = s
                    break
        if signal is None:
            return

        rois = self._get_rois(signal)
        self.updated[BaseSignal, list].emit(signal, rois)
        self.updated[BaseSignal, list, SignalFigureTool].emit(
            signal, rois, self)

    def _cancel(self, signal):
        for w in self.widgets[signal]:
            if w.is_on():
                w.set_on(False)
        self.widgets.pop(signal)
        self.axes.pop(signal)

    def accept(self, signal):
        if self.have_selection(signal):
            rois = self._get_rois(signal)
            self.accepted[BaseSignal, list].emit(signal, rois)
            self.accepted[BaseSignal, list, SignalFigureTool].emit(
                signal, rois, self)
        if signal is not None:
            self.cancel(signal)

    def cancel(self, signal=None):
        if signal is None:
            signals = list(self.widgets.keys())
        else:
            signals = [signal]
        for s in signals:
            if s in self.widgets:
                self._cancel(s)
                self.cancelled[BaseSignal].emit(s)

    def disconnect_windows(self, windows):
        super(MultiSelectionTool, self).disconnect_windows(windows)
        self.cancel()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Mar 01 18:26:38 2015

@author: Vidar Tonaas Fauske
"""

from functools import partial

from qtpy import QtCore
from qtpy.QtCore import QSettings
from qtpy.QtWidgets import (QDialogButtonBox, QVBoxLayout, QLineEdit, QWidget,
                            QCheckBox, QDoubleSpinBox, QSpinBox, QComboBox,
                            QMessageBox, QAbstractButton,
                            QTabWidget, QFormLayout)

from .extendedqwidgets import ExToolWindow
from hyperspyui.settings import Settings

import numpy as np


def tr(text):
    return QtCore.QCoreApplication.translate("PluginManagerWidget", text)


class SettingsDialog(ExToolWindow):
    settings_changed = QtCore.Signal(dict)

    def __init__(self, main_window, parent=None):
        """
        Create a dialog for editing the application settings, including the
        settings for plugins.
        """
        super(SettingsDialog, self).__init__(parent)

        self.setWindowTitle(tr("Settings"))
        self.ui = main_window
        self._initial_values = {}
        self._changes = {}
        self._lut = {}
        self.create_controls()

    @property
    def apply_btn(self):
        """
        The apply button.
        """
        return self._btns.button(QDialogButtonBox.Apply)

    def _on_setting_changed(self, key, widget, *pysideargs):
        """
        Callback when the value of a settings editor widget has changed.
        """
        # First, extract value from widget (depends on widget type)
        if isinstance(widget, QLineEdit):
            v = widget.text()
        elif isinstance(widget, QCheckBox):
            if widget.isTristate() and \
                    widget.checkState() == QtCore.Qt.PartiallyChecked:
                v = None
            else:
                v = "true" if widget.isChecked() else "false"
        elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            v = widget.value()
        elif isinstance(widget, QComboBox):
            v = widget.currentText()

        # Compare to initial value:
        if v == self._initial_values[key]:
            # If the same, remove from self._changes
            del self._changes[key]
            if len(self._changes) < 1:
                # If no changes, disable apply button (nothing to apply)
                self.apply_btn.setEnabled(False)
        else:
            # If different, store in self._changes and enable apply button
            self._changes[key] = v
            self.apply_btn.setEnabled(True)

    def apply_changes(self):
        """
        Applies changes performed since dialog creation or last apply,
        whichever is most recent. Fires settings_changed Signal as long as
        there has been any changes.
        """
        if len(self._changes) < 1:
            return
        s = QSettings(self.ui)
        for k, v in self._changes.items():
            if k in self._initial_values:   # Avoid readding removed settings
                s.setValue(k, v)
                self._initial_values[k] = v
        self.settings_changed.emit(self._changes)
        self._changes.clear()
        self.apply_btn.setEnabled(False)

    def _create_settings_widgets(self, settings):
        """
        Create a widget for a settings instance, containing label/editor pairs
        for each setting in the current level of the passed QSettings instance.
        The key of the setting is used as the label, but it's capitalized and
        underscores are replaced by spaces.
        """
        wrap = QWidget(self)
        form = QFormLayout()
        hint_lookup = Settings()
        for k in settings.allKeys():
            if k.startswith("_"):
                continue                                # Ignore hidden keys
            v = settings.value(k)                       # Read value
            label = k.capitalize().replace('_', ' ')
            abs_key = settings.group() + '/' + k
            self._initial_values[abs_key] = v           # Store initial value
            hints = hint_lookup.get_enum_hint(abs_key)  # Check for enum hints
            # Create a fitting editor widget based on value type:
            if hints is not None:
                w = QComboBox()
                w.addItems(hints)
                w.setEditable(True)
                w.setEditText(v)
                w.editTextChanged.connect(partial(self._on_setting_changed,
                                                  abs_key, w))
            elif isinstance(v, str):
                if v.lower() in ('true', 'false'):
                    w = QCheckBox()
                    w.setChecked(v.lower() == 'true')
                    w.toggled.connect(partial(self._on_setting_changed,
                                              abs_key, w))
                else:
                    w = QLineEdit(v)
                    w.textChanged.connect(partial(self._on_setting_changed,
                                                  abs_key, w))
            elif isinstance(v, int):
                w = QSpinBox()
                w.setRange(np.iinfo(np.int32).min, np.iinfo(np.int32).max)
                w.setValue(v)
                w.valueChanged.connect(partial(self._on_setting_changed,
                                               abs_key, w))
            elif isinstance(v, float):
                w = QDoubleSpinBox()
                w.setRange(np.finfo(np.float32).min, np.finfo(np.float32).max)
                w.setValue(v)
                w.valueChanged.connect(partial(self._on_setting_changed,
                                               abs_key, w))
            else:
                w = QLineEdit(str(v))
                w.textChanged.connect(partial(self._on_setting_changed,
                                              abs_key, w))
            self._lut[abs_key] = w
            form.addRow(label, w)
        wrap.setLayout(form)
        return wrap

    def _add_groups(self, settings):
        """
        Add all child groups in settings as a separate tab, with editor widgets
        to change the values of each setting within those groups.

        Treats the groups 'PluginManager' and 'plugins' specially: The former
        is ignored in its entirety, the latter is called recursively so that
        each plugin gets its own tab.
        """
        for group in settings.childGroups():
            if group in ('defaults', 'PluginManager'):
                continue
            elif group == 'plugins':
                settings.beginGroup(group)
                self._add_groups(settings)
                settings.endGroup()
                continue
            settings.beginGroup(group)
            tab = self._create_settings_widgets(settings)
            settings.endGroup()
            if group.lower() == 'general':
                self.general_tab = tab
                self.tabs.insertTab(0, tab, tr("General"))
            else:
                self.tabs.addTab(tab, group)

    def _on_accept(self):
        """
        Callback when dialog is closed by OK-button.
        """
        self.apply_changes()
        self.accept()

    def _on_reset(self):
        """
        Callback for reset button. Prompts user for confirmation, then proceeds
        to reset settings to default values if confirmed, before updating
        controls and applying any changes (emits change signal if any changes).
        """
        mb = QMessageBox(QMessageBox.Warning,
                         tr("Reset all settings"),
                         tr("This will reset all settings to their default " +
                            "values. Are you sure you want to continue?"),
                         QMessageBox.Yes | QMessageBox.No)
        mb.setDefaultButton(QMessageBox.No)
        dr = mb.exec_()
        if dr == QMessageBox.Yes:
            # This clears all settings, and recreates only those values
            # initialized with set_default this session.
            Settings.restore_from_defaults()

            # Now we update controls:
            s = QSettings(self.ui)
            keys = list(self._initial_values.keys())  # Use copy, as we may modify
            for k in keys:
                # Check if setting is still present
                if s.contains(k):
                    # Present, update to new value (triggers _on_change)
                    v = s.value(k)
                    w = self._lut[k]
                    if isinstance(w, QLineEdit):
                        w.setText(v)
                    elif isinstance(w, QCheckBox):
                        w.setChecked(v.lower() == "true")
                    elif isinstance(w, (QSpinBox, QDoubleSpinBox)):
                        w.setValue(v)
                else:
                    # Setting was removed, remove editor
                    w = self._lut[k]
                    layout = w.parent().layout()
                    label = layout.labelForField(w)
                    layout.removeWidget(w)
                    w.close()
                    if label is not None:
                        layout.removeWidget(label)
                        label.close()
                    del self._lut[k]
                    del self._initial_values[k]
                    self._changes[k] = None
                    # Check whether all editors for tab was removed
                    if layout.count() == 0:
                        wrap = w.parent()
                        self.tabs.removeTab(self.tabs.indexOf(wrap))
            # Finally apply changes (update _initial_values, and emit signal)
            self.apply_changes()

    def _on_click(self, button):
        """
        Route button clicks to appropriate handler.
        """
        if button == self.apply_btn:
            self.apply_changes()
        elif button == self._btns.button(QDialogButtonBox.Reset):
            self._on_reset()

    def create_controls(self):
        """
        Create UI controls.
        """
        self.tabs = QTabWidget(self)

        # Fill in tabs by setting groups
        s = QSettings(self.ui)
        self._add_groups(s)

        # Add button bar at end
        btns = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Apply |
                                QDialogButtonBox.Cancel |
                                QDialogButtonBox.Reset,
                                QtCore.Qt.Horizontal, self)
        btns.accepted.connect(self._on_accept)
        btns.rejected.connect(self.reject)
        btns.clicked[QAbstractButton].connect(self._on_click)
        self._btns = btns
        self.apply_btn.setEnabled(False)

        vbox = QVBoxLayout()
        vbox.addWidget(self.tabs)
        vbox.addWidget(btns)

        self.setLayout(vbox)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Nov 17 14:16:41 2014

@author: Vidar Tonaas Fauske
"""


from qtpy import QtCore, QtWidgets

import types

from hyperspyui.exceptions import ProcessCanceled


def tr(text):
    return QtCore.QCoreApplication.translate("Threaded", text)


class Worker(QtCore.QObject):
    progress = QtCore.Signal(int)
    finished = QtCore.Signal()
    error = QtCore.Signal(str)

    def __init__(self, run):
        super(Worker, self).__init__()
        self.run_function = run

    def process(self):
        self.progress.emit(0)
        try:
            if isinstance(self.run_function, types.GeneratorType):
                p = 0
                self.progress.emit(p)
                while p < 100:
                    p = self.run_function()
                    self.progress.emit(p)
            else:
                self.run_function()
            self.progress.emit(100)
            self.finished.emit()
        except Exception as e:
            self.error.emit(str(e))
            raise


class Threaded(QtCore.QObject):

    """
    Executes a user provided function in a new thread, and pops up a
    QProgressBar until it finishes. To have an updating progress bar,
    have the provided function be a generator, and yield completion rate
    in percent (int from 0 to 100).
    """

    pool = []

    def add_to_pool(instance):
        Threaded.pool.append(instance)

    def remove_from_pool(instance):
        Threaded.pool.remove(instance)

    def __init__(self, parent, run, finished=None):
        super(Threaded, self).__init__(parent)

        # Create thread/objects
        self.thread = QtCore.QThread()
        worker = Worker(run)
        worker.moveToThread(self.thread)
        Threaded.add_to_pool(self)

        # Connect error reporting
        worker.error[str].connect(self.errorString)

        # Start up
        self.thread.started.connect(worker.process)

        # Clean up
        worker.finished.connect(self.thread.quit)
        worker.finished.connect(worker.deleteLater)
        worker.error.connect(self.thread.quit)
        worker.error.connect(worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)

        if finished is not None:
            worker.finished.connect(finished)

        def remove_ref():
            Threaded.remove_from_pool(self)
        self.thread.finished.connect(remove_ref)

        # Need to keep ref so they stay in mem
        self.worker = worker

    def errorString(self, error):
        print(error)

    def run(self):
        self.thread.start()


class ProgressThreaded(Threaded):

    def __init__(self, parent, run, finished=None, label=None, cancellable=False,
                 title=tr("Processing"), modal=True, generator_N=None):
        self.modal = modal
        self.generator_N = generator_N

        # Create progress bar.
        progressbar = QtWidgets.QProgressDialog(parent)
        if isinstance(run, types.GeneratorType):
            progressbar.setMinimum(0)
            if generator_N is None:
                generator_N = 100
            elif generator_N <= 1:
                progressbar.setMaximum(0)
            else:
                progressbar.setMaximum(generator_N)
        else:
            progressbar.setMinimum(0)
            progressbar.setMaximum(0)

#        progressbar.hide()
        progressbar.setWindowTitle(title)
        progressbar.setLabelText(label)
        if not cancellable:
            progressbar.setCancelButtonText(None)

        if isinstance(run, types.GeneratorType):
            def run_gen():
                for p in run:
                    self.worker.progress[int].emit(p)
                    if self.progressbar.wasCanceled():
                        raise ProcessCanceled(tr("User cancelled operation"))

            super(ProgressThreaded, self).__init__(parent, run_gen, finished)
        else:
            super(ProgressThreaded, self).__init__(parent, run, finished)

        self.thread.started.connect(self.display)
        self.worker.finished.connect(self.close)
        self.worker.error.connect(self.close)
        self.worker.progress[int].connect(progressbar.setValue)
        self.progressbar = progressbar

    def display(self):
        if not self.thread.isFinished():
            if self.modal:
                self.progressbar.exec_()
            else:
                self.progressbar.show()

    def close(self):
        self.progressbar.close()
        self.progressbar.deleteLater()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.


from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtWidgets import QDialogButtonBox

from hyperspyui.widgets.extendedqwidgets import ExToolWindow


class AxesPickerDialog(ExToolWindow):
    def __init__(self, ui, signal, single=False):
        super().__init__(ui)
        self.ui = ui
        self.single = single
        self.signal = signal
        self.create_controls()
        self.setWindowTitle("Select axes")

    @property
    def selected_axes(self):
        sel = self.list.selectedItems()
        if self.single:
            if len(sel) == 0:
                return None
            elif len(sel) == 1:
                return sel[0].data(QtCore.Qt.UserRole)
            else:
                raise ValueError("Invalid selection")
        else:
            return [i.data(QtCore.Qt.UserRole) for i in sel]

    def create_controls(self):
        self.list = QtWidgets.QListWidget()
        for ax in self.signal.axes_manager._get_axes_in_natural_order():
            rep = '%s axis, size: %i' % (ax._get_name(), ax.size)
            item = QtWidgets.QListWidgetItem(rep, self.list)
            item.setData(QtCore.Qt.UserRole, ax)
            self.list.addItem(item)
        if not self.single:
            self.list.setSelectionMode(
                QtGui.QAbstractItemView.ExtendedSelection)
        btns = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal)

        btns.accepted.connect(self.accept)
        btns.rejected.connect(self.reject)
        vbox = QtGui.QVBoxLayout(self)
        vbox.addWidget(self.list)
        vbox.addWidget(btns)

        self.setLayout(vbox)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

from hyperspyui.plugins.plugin import Plugin
import matplotlib.pyplot as plt


class Tightlayout(Plugin):
    name = "TightLayout"

    def create_actions(self):
        self.add_action(
            self.name + '.default', "Tight layout", self.default,
            icon="move.svg",
            tip="Apply a tight layout to the selected plot.")

    def create_menu(self):
        self.add_menuitem('Plot', self.ui.actions[self.name + '.default'])

    def create_toolbars(self):
        self.add_toolbar_button(
            'Plot',
            self.ui.actions[
                self.name +
                '.default'])

    def default(self):
        ui = self.ui
        f = plt.gcf()
        if f:
            f.tight_layout()
            f.canvas.draw()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Apr 12 17:43:47 2015

@author: Vidar Tonaas Fauske
"""

import sys
import os
import platform
import subprocess

from hyperspy.io_plugins import io_plugins



def run_desktop_integration_windows(args):
    try:
        create_shortcut
    except NameError:
        # Create a function with the same signature as create_shortcut provided
        # by bdist_wininst
        def create_shortcut(path, description, filename,
                            arguments="", workdir="", iconpath="",
                            iconindex=0):
            with open(filename, 'a'):
                pass    # Touch
            import pythoncom
            from win32com.shell import shell

            ilink = pythoncom.CoCreateInstance(
                shell.CLSID_ShellLink, None,
                pythoncom.CLSCTX_INPROC_SERVER, shell.IID_IShellLink)
            ilink.SetPath(path)
            ilink.SetDescription(description)
            if arguments:
                ilink.SetArguments(arguments)
            if workdir:
                ilink.SetWorkingDirectory(workdir)
            if iconpath or iconindex:
                ilink.SetIconLocation(iconpath, iconindex)
            # now save it.
            ipf = ilink.QueryInterface(pythoncom.IID_IPersistFile)
            ipf.Save(filename, 0)

        # Support the same list of "path names" as bdist_wininst.
        def get_special_folder_path(path_name):
            from win32com.shell import shell, shellcon

            for maybe in """
                    CSIDL_COMMON_STARTMENU CSIDL_STARTMENU CSIDL_COMMON_APPDATA
                    CSIDL_LOCAL_APPDATA CSIDL_APPDATA CSIDL_PROGRAM_FILES
                    CSIDL_COMMON_DESKTOPDIRECTORY CSIDL_DESKTOPDIRECTORY
                    CSIDL_COMMON_STARTUP CSIDL_STARTUP CSIDL_COMMON_PROGRAMS
                    CSIDL_PROGRAMS CSIDL_PROGRAM_FILES_COMMON CSIDL_FONTS
                    """.split():
                if maybe == path_name:
                    csidl = getattr(shellcon, maybe)
                    return shell.SHGetSpecialFolderPath(0, csidl, False)
            raise ValueError("%s is an unknown path ID" % (path_name,))

    pyw_executable = os.path.join(sys.prefix, "pythonw.exe")

    shortcut_filename = "HyperSpyUI.lnk"
    import hyperspyui
    dirname = os.path.dirname(hyperspyui.__file__)
    script_path = os.path.join(dirname, "__main__.py")
    icon_path = os.path.join(dirname, 'images', 'icon', 'hyperspy.ico')

    if args.remove:
        pass    # Should we delete registry entries? Maybe if not edited?
    else:
        if not args.no_shortcuts:
            # Get paths to the desktop and start menu
            print('Creating Shortcuts')
            try:
                desktop_path = get_special_folder_path(
                    "CSIDL_COMMON_DESKTOPDIRECTORY")
                startmenu_path = get_special_folder_path(
                    "CSIDL_COMMON_STARTMENU")

                # Create shortcuts.
                for path in [desktop_path, startmenu_path]:
                    create_shortcut(pyw_executable,
                                    "A Graphical interface for HyperSpy",
                                    os.path.join(path, shortcut_filename),
                                    '"%s"' % script_path,
                                    sys.prefix,
                                    icon_path)
            except IOError as e:
                # Try again with user folders
                desktop_path = get_special_folder_path(
                    "CSIDL_DESKTOPDIRECTORY")
                startmenu_path = get_special_folder_path("CSIDL_STARTMENU")

                # Create shortcuts.
                for path in [desktop_path, startmenu_path]:
                    create_shortcut(pyw_executable,
                                    "A Graphical interface for HyperSpy",
                                    os.path.join(path, shortcut_filename),
                                    '"%s"' % script_path,
                                    sys.prefix,
                                    icon_path)
        
        exclude_formats = [
            "netCDF", # old HyperSpy format.
            "Signal2D", # don't register it to open standard images
            "Protochips", # extension is csv
            ]
        filetypes = []
        for hspy_format in io_plugins:
            if hspy_format.format_name not in exclude_formats:
                defext = hspy_format.file_extensions[hspy_format.default_extension]
                filetypes.append("." + defext)

        if filetypes:
            d = dirname
            docname = "HyperSpy.Document"

            # Setup default icon
            cmd = r'1>nul 2>nul REG ADD "HKCR\%s\DefaultIcon" ' % docname
            cmd += '/t REG_SZ /f /d "' + d + r'\images\icon\hyperspy.ico"'

            # Try to register for all users (requires admin)
            r = subprocess.call(cmd, shell=True)
            cmds = []
            if r == 0:  # Everything OK, we're admin. Use ASSOC and FTYPE
                for ft in filetypes:
                    cmds.append(r'1>nul 2>nul ASSOC %s=%s' % (ft, docname))
                cmds.append(r'1>nul 2>nul FTYPE ' +
                            r'{0}="{1}" "'.format(docname, pyw_executable) +
                            d + r'\__main__.py" "%1" %*')
            else:
                # Not admin. We have to add everything to HKCU manually
                cmd = (r'1>nul 2>nul REG ADD ' +
                       r'"HKCU\Software\Classes\%s\DefaultIcon"' % docname) + \
                       r' /t REG_SZ /f /d "'
                cmd += d + r'\images\icon\hyperspy.ico"'
                cmds.append(cmd)
                for ft in filetypes:
                    cmds.append(
                        (r'1>nul 2>nul REG ADD "HKCU\Software\Classes\%s" ' %
                         ft) +
                        (r'/v "" /t REG_SZ /d "%s" /f' % docname))
                cmds.append(
                    r'1>nul 2>nul REG ADD ' +
                    (r'"HKCU\Software\Classes\%s\shell\open\command"' %
                     docname) +
                    r' /v "" /t REG_EXPAND_SZ /d ' +
                    (r'"\"%s\" "' % pyw_executable) +
                    d + r'\__main__.py" \"%1\" %*" /f')

            for cmd in cmds:
                r = subprocess.call(cmd, shell=True)
            print("File types registered: ", filetypes)



import os
import sys
from subprocess import Popen, PIPE, run, TimeoutExpired

import pytest


here = os.path.abspath(__file__)
app_module = os.path.join(os.path.dirname(here), 'single_app.py')


def wait(proc, timeout):
    try:
        proc.wait(timeout)
    except TimeoutExpired:
        return True
    return False


@pytest.mark.timeout(10)
def test_single_application(request):
    cmd = [sys.executable, '-u', app_module, 'my_flag']
    # Start primary instance
    primary = Popen(cmd, stdout=PIPE, stderr=PIPE,
                    universal_newlines=True)

    def _term():
        try:
            primary.kill()
        except OSError:
            pass
    request.addfinalizer(_term)

    # Wait for completed start-up:
    primary_output = ''
    i = 0
    while wait(primary, 1):
        i += 1
        primary_output += primary.stdout.readline()
        if primary_output:
            break
        elif i >= 5:
            raise TimeoutError(
                'Timed out waiting for primary start.'
                '\n\tstdout: %s\n\tstderr: %s' % (
                    primary_output, primary.stderr.read())
                )
    if primary.poll() is not None:
        raise ValueError(
            'Primary instance exited prematurely!'
            '\n\tstdout: %s\n\tstderr: %s' % (
                primary_output, primary.stderr.read()))

    # Start secondary instance
    secondary = run(cmd, timeout=5)

    # Check that secondary instance exited as expected
    assert secondary.returncode == 1

    # Check that primary instance exited as expected
    out, err = primary.communicate(timeout=5)
    primary_output += out
    assert primary.returncode == 0
    assert 'my_flag' in primary_output



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Dec 07 02:03:23 2014

@author: Vidar Tonaas Fauske
"""

import os

from .figuretool import FigureTool
from hyperspyui.util import load_cursor


class ZoomPanTool(FigureTool):

    def __init__(self, windows=None):
        super(ZoomPanTool, self).__init__(windows)
        self.panning = False
        self.pan_data = None
        self.base_scale = 1.5     # Mouse wheel zoom factor

    def get_name(self):
        return "Pan/Zoom tool"

    def get_category(self):
        return 'Plot'

    def get_icon(self):
        return os.path.dirname(__file__) + '/../images/panzoom2.svg'

    def is_selectable(self):
        return True

    def make_cursor(self):
        return load_cursor(os.path.dirname(__file__) +
                           '/../images/panzoom2.svg', 8, 8)

    def on_mousedown(self, event):
        if event.inaxes is None:
            return
        f = event.inaxes.figure
        x, y = event.x, event.y
        axes = []
        for a in f.get_axes():
            if (x is not None and y is not None and a.in_axes(event) and
                    a.can_pan()):
                a.start_pan(x, y, event.button)
                axes.append(a)
        if len(axes) > 0:
            self.panning = True
            self.pan_data = [x, y, event.button, axes]

    def on_mouseup(self, event):
        if not self.panning:
            return
        self.panning = False
        if self.pan_data is None:
            return
        for a in self.pan_data[3]:
            a.end_pan()
            a.figure.canvas.draw_idle()
        self.pan_data = None

    def on_mousemove(self, event):
        if not self.panning or self.pan_data is None:
            return

        x0, y0 = self.pan_data[0:2]
        x1, y1 = self.pan_data[0:2] = (event.x, event.y)

        cs = set()
        for a in self.pan_data[3]:
            # safer to use the recorded button at the press than current button
            # as multiple buttons can get pressed during motion...
            a.drag_pan(self.pan_data[2], event.key, event.x, event.y)
            cs.add(a.figure.canvas)
        for canvas in cs:
            canvas.draw()

    def connect_windows(self, windows):
        super(ZoomPanTool, self).connect_windows(windows)
        windows = self._iter_windows(windows)
        canvases = set()
        for w in windows:
            canvases.add(w.widget())
        for c in canvases:
            c.widgetlock(self)

    def disconnect_windows(self, windows):
        super(ZoomPanTool, self).disconnect_windows(windows)
        windows = self._iter_windows(windows)
        canvases = set()
        for w in windows:
            canvases.add(w.widget())
        for c in canvases:
            c.widgetlock.release(self)

    def on_scroll(self, event):
        if event.inaxes is None:
            return
        ax = event.inaxes
        # get the current x and y limits
        cxlim = ax.get_xlim()
        cylim = ax.get_ylim()
        cx = (cxlim[1] - cxlim[0]) * .5
        cy = (cylim[1] - cylim[0]) * .5
        x = event.xdata  # get event x location
        y = event.ydata  # get event y location
        center = ((cxlim[1] + cxlim[0]) * .5, (cylim[1] + cylim[0]) * .5)
        zoom_vector = (
            (x - center[0]) / self.base_scale,
            (y - center[1]) / self.base_scale)
        if event.button == 'up':
            # deal with zoom in
            scale = 1.0 / self.base_scale
            new_centre = (
                center[0] + zoom_vector[0], center[1] + zoom_vector[1])
            new_xlim = [new_centre[0] - cx * scale,
                        new_centre[0] + cx * scale]
            new_ylim = [new_centre[1] - cy * scale,
                        new_centre[1] + cy * scale]
        elif event.button == 'down':
            # deal with zoom out
            scale = self.base_scale
            new_centre = (
                center[0] - zoom_vector[0], center[1] - zoom_vector[1])
            new_xlim = [new_centre[0] - cx * scale,
                        new_centre[0] + cx * scale]
            new_ylim = [new_centre[1] - cy * scale,
                        new_centre[1] + cy * scale]
        else:
            return
        # set new limits
        ax.set_xlim(new_xlim)
        ax.set_ylim(new_ylim)
        ax.figure.canvas.draw_idle()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Apr 17 00:23:43 2015

@author: Vidar Tonaas Fauske
"""

from hyperspyui.plugins.plugin import Plugin

import os
from scipy import stats
import numpy as np

from hyperspyui.tools import SelectionTool
from hyperspy.roi import BaseInteractiveROI
from hyperspy.utils.markers import line_segment, text


class FittingPlugin(Plugin):
    name = "Fitting plugin"

    def create_tools(self):
        self.reg_tool = RegressionTool()
        self.reg_tool.category = 'Spectrum'
        self.reg_tool.accepted[BaseInteractiveROI].connect(self.regression)
        self.add_tool(self.reg_tool, self.ui.select_signal)

    def regression(self, roi, signal=None, axes=None):
        if signal is None:
            f = self.reg_tool.widget.ax.figure
            window = f.canvas.parent()
            sw = window.property('hyperspyUI.SignalWrapper')
            if sw is None:
                return
            signal = sw.signal
        sig_axes = signal.axes_manager._axes
        if axes is None:
            axes = self.reg_tool.axes
        else:
            axes = sig_axes[axes]

        i_ax = sig_axes.index(axes[0])
        slices = list(roi._make_slices(sig_axes, axes))
        for i, a in enumerate(sig_axes):
            if i != i_ax:
                slices[i] = a.index

        y = signal.data[tuple(slices)]
        x = axes[0].axis[slices[i_ax]]

        # TODO: If in signal dim, iterate through navigation space
        reg = stats.linregress(x, y)
        x1, x2 = np.min(x), np.max(x)
        y1, y2 = np.array([x1, x2]) * reg[0] + reg[1]
        m_l = line_segment(x1, y1, x2, y2)
        signal.add_marker(m_l)
        m_t = text((x2+x1)*0.5, (y2+y1)*0.5,
                   "y = %.4gx + %.4g, R=%.4g" % reg[0:3])
        signal.add_marker(m_t)

        self.record_code("signal = ui.get_selected_signal()")
        self.record_code("axes = " +
                         str(tuple([sig_axes.index(a) for a in axes])))
        self.record_code("roi = utils.roi." + str(roi))
        self.record_code("<p>.regression(roi, signal, axes)")
        self.reg_tool.cancel()   # Turn off functionality as we are finished


class RegressionTool(SelectionTool):

    """
    """

    def __init__(self, *args, **kwargs):
        super(RegressionTool, self).__init__(*args, **kwargs)
        self.valid_dimensions = [1]

    def on_keyup(self, event):
        if event.key == 'delete':   # TODO: OR escape + inactive
            # TODO: Delete regression
            pass
        else:
            super(RegressionTool, self).on_keyup(event)

    def get_name(self):
        return "Regression tool"

    def get_icon(self):
        return os.path.dirname(__file__) + '/../images/regression.svg'



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Dec 07 01:48:00 2014

@author: Vidar Tonaas Fauske
"""

import collections
from qtpy import QtCore

from .tool import Tool


class FigureTool(Tool):

    def __init__(self, windows=None):
        super(FigureTool, self).__init__()
        self.cids = {}
        self.cursor = self.make_cursor()
        if self.single_action() is not None:
            self.connect_windows(windows)

    def make_cursor(self):
        """
        Initialize the cursor for this instance. Is called in constructor.
        """
        return QtCore.Qt.ArrowCursor

    def get_cursor(self, widget=None):
        """
        Get the cursor for the supplied widget. Is applied when a widget is
        connected to the Tool.
        """
        # This default implementation does not use the widget information for
        # anything, but descendants might.
        return self.cursor

    def get_window(self, event):
        """
        Get the window for the event
        """
        if event.insaxes is None:
            return None
        return event.inaxes.figure.canvas.parent()

    def get_pixel_size(self, event):
        """
        Get the point size in data units
        """
        ax = event.inaxes
        if ax is None:
            return (0, 0)
        invtrans = self.ax.transData.inverted()
        return abs(invtrans.transform((1, 1)) -
                   invtrans.transform((0, 0)))

    def _wire(self, canvas, local_key, mpl_key):
        """Connect an MPL event to an instance method, if the method is defined
        on the current instance. The local method is defined by 'local_key'.
        """
        if hasattr(self, local_key):
            if local_key not in self.cids:
                self.cids[local_key] = {}
            self.cids[local_key][canvas] = canvas.mpl_connect(
                mpl_key, getattr(self, local_key))

    @staticmethod
    def _iter_windows(windows):
        if windows is None:
            return ()
        if not isinstance(windows, collections.abc.Iterable):
            windows = (windows,)
        return windows

    def connect_windows(self, windows):
        """Connects the tool to the windows that are passed. This means that it
        connects to the appropriate MPL events of each window.
        """
        windows = self._iter_windows(windows)
        for w in windows:
            canvas = w.widget()
            canvas.setCursor(self.get_cursor(canvas))
            self._wire(canvas, 'on_mousemove', 'motion_notify_event')
            self._wire(canvas, 'on_mousedown', 'button_press_event')
            self._wire(canvas, 'on_mouseup', 'button_release_event')
            self._wire(canvas, 'on_keydown', 'key_press_event')
            self._wire(canvas, 'on_keyup', 'key_release_event')
            self._wire(canvas, 'on_pick', 'pick_event')
            self._wire(canvas, 'on_scroll', 'scroll_event')
            self._wire(canvas, 'on_draw', 'draw_event')
            self._wire(canvas, 'on_resize', 'resize_event')
            self._wire(canvas, 'on_figure_enter', 'figure_enter_event')
            self._wire(canvas, 'on_figure_leave', 'figure_leave_event')
            self._wire(canvas, 'on_axes_enter', 'axes_enter_event')
            self._wire(canvas, 'on_axes_leave', 'axes_leave_event')

    def disconnect_windows(self, windows):
        windows = self._iter_windows(windows)
        for w in windows:
            canvas = w.widget()
            for cid_iter in self.cids.values():
                for canv, cid in cid_iter.items():
                    if canv == canvas:
                        try:
                            canvas.mpl_disconnect(cid)
                        except:
                            pass



import sys
import argparse

parser = argparse.ArgumentParser(
    description='Install HyperspyUI into the OS')
parser.add_argument('--remove', action="store_true", help='uninstall flag')
parser.add_argument('--no-shortcuts', action="store_true",
                    help='(Windows only) do not create shortcuts')

args = parser.parse_args()


if __name__ == "__main__":
    if sys.platform.startswith("linux"):
        from hyperspyui.desktop_integration_linux import run_desktop_integration_linux
        run_desktop_integration_linux(args)
    elif sys.platform == 'win32':
        from hyperspyui.desktop_integration_windows import run_desktop_integration_windows
        run_desktop_integration_windows(args)
    else:
        raise NotImplementedError(
            "Desktop integration is currently only available for Linux and Microsoft Windows. Current OS is {}.".format( sys.platform))


# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sun Dec 07 02:03:23 2014

@author: Vidar Tonaas Fauske
"""

from hyperspyui.plugins.plugin import Plugin

import os

from hyperspy.roi import BaseInteractiveROI
from hyperspyui.tools import SelectionTool
from hyperspyui.util import load_cursor


class CropToolPlugin(Plugin):
    name = "Crop tool"

    def create_tools(self):
        self.tool = CropTool()
        self.tool.accepted[BaseInteractiveROI].connect(self.crop)
        self.add_tool(self.tool, self.ui.select_signal)

    def crop(self, roi, signal=None, axes=None):
        if signal is None:
            f = self.tool.widget.ax.figure
            window = f.canvas.parent()
            sw = window.property('hyperspyUI.SignalWrapper')
            if sw is None:
                return
            signal = sw.signal
        sig_axes = signal.axes_manager._axes
        if axes is None:
            axes = self.tool.axes
        else:
            old_axes = axes
            axes = []
            for a in old_axes:
                axes.append(sig_axes[a])
        slices = roi._make_slices(sig_axes, axes)
        signal.data = signal.data[slices]
        if roi.ndim > 0:
            signal.axes_manager[axes[0]].offset = roi.left
        if roi.ndim > 1:
            signal.axes_manager[axes[1]].offset = roi.top

        signal.squeeze()
        signal.get_dimensions_from_data()
        signal.events.data_changed.trigger(signal)

        self.record_code("s_crop = ui.get_selected_signal()")
        self.record_code("axes = " +
                         str(tuple([sig_axes.index(a) for a in axes])))
        self.record_code("roi = hs.roi." + str(roi))
        self.record_code("<p>.crop(roi, s_crop, axes)")
        self.tool.cancel()   # Turn off functionality as we are finished


class CropTool(SelectionTool):

    """
    Tool to crop signal interactively. Simply click and drag in a figure to
    create an ROI, and then press enter to apply cropping operation, or ESC to
    abort cropping. The cropping can also be aborted simply by selecting a
    different tool.
    """

    def get_name(self):
        return "Crop tool"

    def get_icon(self):
        return os.path.dirname(__file__) + '/../images/crop.svg'

    def make_cursor(self):
        return load_cursor(os.path.dirname(__file__) +
                           '/../images/crop.svg', 8, 8)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Apr 17 00:03:50 2015

@author: Vidar Tonaas Fauske
"""

from qtpy import QtCore

from hyperspy.drawing.widgets import (RectangleWidget, RangeWidget,
                                      SquareWidget, VerticalLineWidget)
from hyperspy.roi import (BaseInteractiveROI, RectangularROI, SpanROI,
    Point1DROI, Point2DROI)

from hyperspyui.tools import SignalFigureTool
from hyperspyui.util import crosshair_cursor


class SelectionTool(SignalFigureTool):

    """
    Tool to select a ROI on a  signal interactively. Simply click and drag in a
    figure to create an ROI, and then press enter to indicate that the
    selection is complete, or ESC to cancel. The selection can also be aborted
    by simply selecting a different tool.
    """

    accepted = QtCore.Signal([BaseInteractiveROI],
                             [BaseInteractiveROI, SignalFigureTool])
    updated = QtCore.Signal([BaseInteractiveROI],
                            [BaseInteractiveROI, SignalFigureTool]) # TODO: Use
    cancelled = QtCore.Signal()

    def __init__(self, windows=None, name=None, category=None, icon=None,
                 description=None):
        super(SelectionTool, self).__init__(windows)
        self.widget2d_r = RectangleWidget(None)
        self.widget2d_r.set_on(False)
        self.widget1d_r = RangeWidget(None)
        self.widget1d_r.set_on(False)
        self.widget2d = SquareWidget(None)
        self.widget2d.set_on(False)
        self.widget1d = VerticalLineWidget(None)
        self.widget1d.set_on(False)
        self._widgets = [self.widget2d_r, self.widget1d_r, self.widget2d,
                         self.widget1d]
        self.valid_dimensions = [1, 2]
        self.ranged = True

        self.name = name or "Selection tool"
        self.category = category or 'Signal'
        self.description = description
        self.icon = icon
        self.cancel_on_accept = False

    @property
    def widget(self):
        if self.ndim == 1:
            if self.ranged:
                return self.widget1d_r
            else:
                return self.widget1d
        else:
            if self.ranged:
                return self.widget2d_r
            else:
                return self.widget2d

    @property
    def ndim(self):
        if self.axes is None:
            return 0
        return len(self.axes)

    def is_on(self):
        return self.widget.is_on()

    def in_ax(self, ax):
        return ax == self.widget.ax

    def get_name(self):
        return self.name

    def get_category(self):
        return self.category

    def get_icon(self):
        return self.icon

    def make_cursor(self):
        return crosshair_cursor()

    def is_selectable(self):
        return True

    def on_mousedown(self, event):
        # Only accept mouse down inside axes
        if event.inaxes is None:
            return

        axes = self._get_axes(event)
        # Make sure we have a figure with valid dimensions
        if len(axes) not in self.valid_dimensions:
            return
        self.axes = axes
        # If we already have a widget, make sure dragging is passed through
        if self.is_on():
            if any([p.contains(event)[0] == True for p in self.widget.patch]):
                return              # Moving, handle in widget
            # Clicked outside existing widget, check for resize handles
            if self.ndim > 1 and self.widget.resizers:
                for r in self.widget._resizer_handles:
                    if r.contains(event)[0] == True:
                        return      # Leave the event to widget
            # Cancel previous and start new
            self.cancel()
            # Cancel reset axes, so set again
            self.axes = axes

        # Find out which axes of Signal are plotted in figure
        s = self._get_signal(event.inaxes.figure)
        if s is None:
            return
        am = s.axes_manager

        self.widget.axes = axes
        if self.ndim > 0:
            self.widget.axes_manager = am

        # Have what we need, create widget
        x, y = event.xdata, event.ydata
        self.widget.set_mpl_ax(event.inaxes)  # connects
        if self.ndim == 1:
            self.widget.position = (x,)
            self.widget.size = axes[0].scale
            self.widget.set_on(True)
            if self.ranged:
                span = self.widget.span
                span.buttonDown = True
                span.on_move_cid = \
                    span.canvas.mpl_connect('motion_notify_event',
                                            span.move_right)
            else:
                self.widget.picked = True
        else:
            self.widget.position = (x, y)
            self.widget.size = [ax.scale for ax in axes]
            if self.ranged:
                self.widget.resizer_picked = 3
            self.widget.picked = True
            self.widget.set_on(True)

    def on_keyup(self, event):
        if event.key == 'enter':
            self.accept()
        elif event.key == 'escape':
            self.cancel()

    def accept(self):
        if self.is_on():
            if self.ndim == 1:
                if self.ranged:
                    roi = SpanROI(0, 1)
                else:
                    roi = Point1DROI(0)
            elif self.ndim > 1:
                if self.ranged:
                    roi = RectangularROI(0, 0, 1, 1)
                else:
                    roi = Point2DROI(0, 0)
            else:
                raise RuntimeWarning("No figure could be found.")
                return
            roi._on_widget_change(self.widget)  # ROI gets coords from widget
            self.accepted[BaseInteractiveROI].emit(roi)
            self.accepted[BaseInteractiveROI, SignalFigureTool].emit(roi, self)
        if self.cancel_on_accept:
            self.cancel()

    def cancel(self):
        for w in self._widgets:
            if w.is_on():
                w.set_on(False)
        self.axes = None
        self.cancelled.emit()

    def disconnect_windows(self, windows):
        super(SelectionTool, self).disconnect_windows(windows)
        self.cancel()
